'''Tests automatic conversion routines.

Copyright William Clarke, University of Oxford 2023
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path

test_base = Path(__file__).parent / 'spec2nii_test_data'
testdata = {
    'twix': test_base / 'Siemens' / 'VEData' / 'Twix' / 'meas_MID00242_FID62747_svs_se_iso_tra_sat.dat',
    'rda': test_base / 'Siemens' / 'XAData' / 'XA20' / 'rda' / 'spct_002.MR.MRI-LAB Test_Dir.5.1.114540.rda',
    'spar': test_base / 'philips' / 'P1' / 'SV_PRESS_sh_6_2_raw_act.SPAR',
    'data_list': test_base / 'philips' / 'hyper' / 'raw_226.list',
    'ge_p': test_base / 'ge' / 'pFiles' / 'svMRS' / 'P03072.7',
    'dicom_siemens': test_base / 'Siemens' / 'XAData' / 'XA20' / 'DICOM' / '26516628.dcm',
    'dicom_uih': test_base / 'UIH' / 'mrs_data' / 'dicom' / 'svs_press_te144_SVS_801' / '00000001.dcm',
    'dicom_philips': test_base / 'philips' / 'DICOM' / 'SV_phantom_center' / 'IM-0018-0002-0001.dcm'}


def test_auto_twix(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['twix'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()


def test_auto_rda(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['rda'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()


def test_auto_sdat_spar(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['spar'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()


def test_auto_ge(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['ge_p'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()


def test_auto_siemens_dicom(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_siemens'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()

    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_siemens'].parent,
        '-o', tmp_path,
        '-f', 'test_dir'])

    assert (tmp_path / 'test_dir.nii.gz').is_file()


def test_auto_uih_dicom(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_uih'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()

    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_uih'].parent,
        '-o', tmp_path,
        '-f', 'test_dir'])

    assert (tmp_path / 'test_dir.nii.gz').is_file()


def test_auto_philips_dicom(tmp_path):
    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_philips'],
        '-o', tmp_path,
        '-f', 'test_file'])

    assert (tmp_path / 'test_file.nii.gz').is_file()

    subprocess.run([
        'spec2nii', 'auto',
        testdata['dicom_philips'].parent,
        '-o', tmp_path,
        '-f', 'test_dir'])

    assert (tmp_path / 'test_dir.nii.gz').is_file()



"""Function to anonymise existing NIfTI-MRS files.

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2021 University of Oxford
"""

import re
import pprint

from nifti_mrs.definitions import standard_defined
from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.nifti_mrs import NIFTI_MRS
from nifti_mrs.hdr_ext import Hdr_Ext


def anon_nifti_mrs(nifti_mrs_img, extra_keys=None, verbose=False):
    """Function for anonymising an existing nifti-mrs object

    :param nifti_mrs_img: NIfTI-MRS object to anonymise
    :type nifti_mrs_img: nifti_mrs.nifti_mrs.NIFTI_MRS
    :param extra_keys: List of keys to explicitly remove, defaults to None
    :type extra_keys: List, optional
    :param verbose: Print verbose output, defaults to False
    :type verbose: bool, optional
    :return: Anonymised NIfTI-MRS object
    :rtype: nifti_mrs.nifti_mrs.NIFTI_MRS
    """

    # Extract header extension
    hdr_ext = nifti_mrs_img.hdr_ext.to_dict()

    # Loop through fields. Remove those which are marked for anonymisation
    # Either those fields which are standard-defined and are marked for anonymisation,
    # or user-defined which are marked 'private'.
    pvt_re = re.compile(r'private_')

    def iter_json(in_dict):
        out_dict = {}
        removed = {}
        for key, value in in_dict.items():
            # Explicitly set on command line
            if extra_keys\
                    and key in extra_keys:
                removed.update({key: value})
            # Standard defined
            elif key in standard_defined:
                if standard_defined[key][3]:
                    removed.update({key: value})
                else:
                    out_dict.update({key: value})
            # User defined
            else:
                # If marked as private at this level
                if pvt_re.match(key):
                    removed.update({key: value})
                # If nested dict
                elif isinstance(value, dict):
                    tmp_out, tmp_rem = iter_json(value)
                    if len(tmp_rem) > 0:
                        removed.update({key: tmp_rem})
                    out_dict.update({key: tmp_out})
                # Catch all other cases
                else:
                    out_dict.update({key: value})
        return out_dict, removed

    anon_hdr, removed_dict = iter_json(hdr_ext)

    if verbose:
        pp = pprint.PrettyPrinter(indent=4)
        print('\nThe following keys were removed:')
        pp.pprint(removed_dict)
        print('\nThe following keys were kept:')
        pp.pprint(anon_hdr)

    # Make new NIfTI-MRS image
    return gen_nifti_mrs_hdr_ext(
        nifti_mrs_img[:],
        nifti_mrs_img.dwelltime,
        Hdr_Ext.from_header_ext(anon_hdr),
        affine=nifti_mrs_img.getAffine('voxel', 'world'))


def anon_file(args):
    """Function for anonymising NIfTI-MRS files stored on drive.

    :param args: Command line arguments parsed in spec2nii.py
    :return: List of anonymised images
    :rtype: [nib.nifti2.Nifti2Image,]
    :return: List of output names
    :rtype: [str,]
    """

    # Process output name.
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [args.file.with_suffix('').with_suffix('').name, ]

    return [
        anon_nifti_mrs(
            NIFTI_MRS(args.file),
            verbose=args.verbose,
            extra_keys=args.remove), ], fname_out



"""Functions to handle sequences/situations that are special-cased
converted from twix/.dat data.

Will Clarke, University of Oxford, 2022
"""
import numpy as np
import re


def smm_svs_herc_hyper(twixObj, reord_data, meta_obj, dim_tags, subseq, subseq_name):
    """
    Special case handling for the Hyper/ISTHMUS Sequence

    The Integrated Short-TE and Hadamard-edited Multi-Sequence (ISTHMUS)
      incorporates a Short TE (35ms) PRESS, Long-TE (80ms) HERCULES, and
      a water reference for each.

    Data is organized into 4 blocks:
      1. Short TE Reference : 35ms Unsuppressed Water
      2. Long  TE Reference : 80ms Unsuppressed Water
      3. Short TE Unedited  : 35ms Water Suppressed PRESS
      4. Long  TE Edited    : 80ms Water Suppressed HERCULES

    Adding these Header Values Here for Reference. Perhaps they'll save someone a headache..
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree',  '3')] = Total Number of Water Refs
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree',  '4')] = Total Number of PRESS
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree',  '5')] = Total Number of HERCULES
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree',  '6')] = Short TE's Echo Time
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree',  '7')] = Sequence Mode (See Below)
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree', '12')] = Pulse Duration

        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree',  '8')] = 4.58 ppm
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree',  '9')] = 1.90 ppm
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '10')] = 3.67 ppm (unused here)
        twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '11')] = 4.18 ppm

    Sequence Mode Referred to Above is Carried Over from mgs_svs sequences:
        0 = MEGA PRESS
        1 = HERMES GABA GSH
        2 = HERMES GABA GSH EtOH
        3 = HERCULES  (This is what should get chosen for HBCD)
        4 = HERMES GABA LAC

    Author : Aaron Gudmundson, Johns Hopkins University, 2023
    Contact: agudmun2@jhmi.edu
    """

    print(f'{subseq:3d} {subseq_name:<20} - Starting   -', end='\r')

    short_TE        = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree', '6')] / 1E6          # Short TE's Echo Time
    pulse_length    = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree', '12')] / 1E6         # Pulse Duration

    edit_cases      = 4                                                                         # 4 Editing Conditions
    edit_pulse_1    = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '8')]                # 4.58 ppm
    edit_pulse_2    = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '9')]                # 1.90 ppm
    edit_pulse_4    = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '11')]               # 4.18 ppm

    dim_info        = f'Hyper - {subseq_name}'                                                  # Subscan Name

    if subseq == 0:                                                                             # Short TE Water Ref
        reord_data = reord_data[:, :, 33::66]                                                   # Short TE Water Refs
        meta_obj.set_standard_def('EchoTime', short_TE)                                         # Echo Time
        meta_obj.set_standard_def('WaterSuppressed', False)                                     # Water Suppression
        meta_obj.set_standard_def('TxOffset', 0.0)                                              # Transmitter Frequency

    elif subseq == 1:                                                                           # Long  TE Water Ref
        reord_data  = reord_data[:, :, 0::66]                                                   # Long  TE Water Refs
        meta_obj.set_standard_def('EchoTime', 0.080)                                            # Echo Time
        meta_obj.set_standard_def('WaterSuppressed', False)                                     # Water Suppression
        meta_obj.set_standard_def('TxOffset', 0.0)                                              # Transmitter Frequency

    elif subseq == 2:                                                                           # Short TE PRESS
        meta_obj.set_standard_def('EchoTime', short_TE)                                         # Echo Time
        meta_obj.set_standard_def('WaterSuppressed', True)                                      # Water Suppression
        meta_obj.set_standard_def('TxOffset', -1.7)                                             # Transmitter Frequency
        reord_data  = reord_data[:, :, 1:33]                                                    # Isolated PRESS

    elif subseq == 3:                                                                           # Long  TE HERCULES
        meta_obj.set_standard_def('EchoTime', 0.080)                                            # Echo Time
        meta_obj.set_standard_def('WaterSuppressed', True)                                      # Water Suppression
        meta_obj.set_standard_def('TxOffset', -1.7)                                             # Transmitter Frequency

        reord_mask       = np.ones(reord_data.shape[-1], dtype=bool)                            # Create a Mask
        reord_mask[::33] = False                                                                # Remove Water Refs
        reord_mask[: 33] = False                                                                # Remove PRESS
        reord_data       = reord_data[..., reord_mask]                                          # Isolated HERCULES

        # Handle Incomplete
        if reord_data.shape[-1] % 4 != 0:
            old_num_avgs = reord_data.shape[-1]                                                 # Old Total Averages
            new_num_avgs = (reord_data.shape[-1] // 4) * 4                                      # New Total Averages
            reord_data   = reord_data[..., :new_num_avgs]                                       # Remove Incomplete

            notestring   = f'{subseq:3d} {subseq_name:<20}'                                     # Note Incomplete Data
            notestring   = f'{notestring} - Correcting - Incomplete Averages'                   # Note Incomplete Data
            notestring   = f'{notestring}  {old_num_avgs} --> {new_num_avgs}'                   # Note Incomplete Data
            print(f'{notestring} \t Corrected**', end='\r')                                     # Note Incomplete Data

        dim_header       = {'EditCondition': ['A', 'B', 'C', 'D']}                              # 4 Subscans
        edit_pulse_val   = {'A': {'PulseOffset': [edit_pulse_1, edit_pulse_2], 'PulseDuration': pulse_length},
                            'B': {'PulseOffset': [edit_pulse_4, edit_pulse_2], 'PulseDuration': pulse_length},
                            'C': {'PulseOffset': edit_pulse_1, 'PulseDuration': pulse_length},
                            'D': {'PulseOffset': edit_pulse_4, 'PulseDuration': pulse_length}}

        meta_obj.set_standard_def('EditPulse', edit_pulse_val)                                  # Header Edit Info

        orig_shape  = reord_data.shape[:-1]                                                     # Remove Averages Dim
        orig_shape += (edit_cases, -1)                                                          # Include Subscans
        reord_data  = reord_data.T.reshape(orig_shape[::-1]).T                                  # With Subscan Dim

        dim_tags.insert(len(orig_shape) - 3, 'DIM_EDIT')                                        # Update Dimensions

    for idx, dt in enumerate(dim_tags):                                                         # Iterate Dimensions
        if dt == 'DIM_EDIT':
            meta_obj.set_dim_info(idx, dt, dim_info, dim_header)                                # Set Dimension
        else:
            meta_obj.set_dim_info(idx, dt)                                                      # Set Dimension

    print(f'{subseq:3d} {subseq_name:<20} - Completed  - Final Array Size: ', reord_data.shape)

    return reord_data, meta_obj, dim_tags


def mgs_svs_ed_twix(twixObj, reord_data, meta_obj, dim_tags):
    """Special case handling for the mgs_svs_ed (VX) and smm_svs_herc (XA) sequence

    MEGA/HERCULES sequence (2/4 editing case)
    """
    seq_mode = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree', '7')]
    pulse_length = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'alFree', '12')] / 1E6
    edit_pulse_1 = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '8')]
    edit_pulse_2 = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '9')]
    edit_pulse_3 = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '10')]
    edit_pulse_off = twixObj['hdr']['Phoenix'][('sWipMemBlock', 'adFree', '11')]

    if seq_mode == 0.0:
        # MEGA-PRESS
        edit_cases = 2
        dim_info = "MEGA-EDITED j-difference editing, two conditions"
        dim_header = {"EditCondition": ["ON", "OFF"]}
        edit_pulse_val = {
            "ON": {"PulseOffset": edit_pulse_1, "PulseDuration": pulse_length},
            "OFF": {"PulseOffset": edit_pulse_off, "PulseDuration": pulse_length}}
    elif seq_mode == 1.0:
        # HERMES GABA GSH (3 edit, 1 ctrl condition)
        edit_cases = 4
        dim_info = "HERMES j-difference editing, GABA GSH, four conditions"
        dim_header = {"EditCondition": ["A", "B", "C", "D"]}
        edit_pulse_val = {
            "A": {"PulseOffset": edit_pulse_1, "PulseDuration": 0.02},
            "B": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02},
            "C": {"PulseOffset": edit_pulse_2, "PulseDuration": 0.02},
            "D": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02}}
    elif seq_mode == 2.0:
        # HERMES GABA GSH EtOH (3 edit, 1 ctrl condition)
        edit_cases = 4
        dim_info = "HERMES j-difference editing, GABA GSH EtOH, four conditions"
        dim_header = {"EditCondition": ["A", "B", "C", "D"]}
        edit_pulse_val = {
            "A": {"PulseOffset": [edit_pulse_1, edit_pulse_3], "PulseDuration": 0.02},
            "B": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02},
            "C": {"PulseOffset": [edit_pulse_2, edit_pulse_3], "PulseDuration": 0.02},
            "D": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02}}
    elif seq_mode == 3.0:
        # HERCULES (4 edit conditions)
        edit_cases = 4
        dim_info = "HERCULES j-difference editing, four conditions"
        dim_header = {"EditCondition": ["A", "B", "C", "D"]}
        edit_pulse_val = {
            "A": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02},
            "B": {"PulseOffset": [edit_pulse_off, edit_pulse_2], "PulseDuration": 0.02},
            "C": {"PulseOffset": edit_pulse_1, "PulseDuration": 0.02},
            "D": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02}}
    elif seq_mode == 4.0:
        # This is possibly only a condition for smm_svs_herc as not present in VE11c version.
        # HERMES GABA LAC (3 edit 1 ctrl conditions)
        edit_cases = 4
        dim_info = "HERMES j-difference editing, GABA LAC, four conditions"
        dim_header = {"EditCondition": ["A", "B", "C", "D"]}
        edit_pulse_val = {
            "A": {"PulseOffset": edit_pulse_1, "PulseDuration": 0.02},
            "B": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02},
            "C": {"PulseOffset": edit_pulse_2, "PulseDuration": 0.02},
            "D": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02}}
    else:
        raise ValueError('Unknown sequence mode in mgs_svs_ed sequence.')

    orig_shape = reord_data.shape[:-1]  # Remove averages dimension
    orig_shape += (edit_cases, -1)
    reord_data = reord_data.T.reshape(orig_shape[::-1]).T

    dim_tags.insert(len(orig_shape) - 3, 'DIM_EDIT')

    # Update the metadata
    for idx, dt in enumerate(dim_tags):
        if dt == 'DIM_EDIT':
            meta_obj.set_dim_info(
                idx,
                dt,
                dim_info,
                dim_header)
        else:
            meta_obj.set_dim_info(idx, dt)

    meta_obj.set_standard_def("EditPulse", edit_pulse_val)

    return reord_data, meta_obj, dim_tags


def slaser_dkd(twixObj, reord_data, meta_obj, dim_tags):

    scan_mode = twixObj.hdr['MeasYaps'][('sSpecPara', 'lAutoRefScanMode')]
    seq_file_name = twixObj['hdr']['Meas'][('tSequenceFileName')].lower()

    group_out = []

    if scan_mode == 1.0:
        for idx, dt in enumerate(dim_tags):
            meta_obj.set_dim_info(idx, dt)
        group_out.append(
            (reord_data.reshape((1, 1, 1) + reord_data.shape),
             meta_obj,
             ''))

    elif (re.search(r'svs_slaser(voi)?_dkd2$', seq_file_name) and scan_mode == 8.0)\
            or (re.search(r'svs_slaser(voi)?_dkd$', seq_file_name) and scan_mode == 8.0):
        num_ref = int(twixObj.hdr['MeasYaps'][('sSpecPara', 'lAutoRefScanNo')])
        num_dyn = int(twixObj.hdr['MeasYaps'][('lAverages',)])

        base_shape = (1, 1, 1) + reord_data.shape[:2] + (-1, )

        for idx, dt in enumerate(dim_tags):
            meta_obj.set_dim_info(idx, dt)

        # Main acq
        start = (num_ref * 2)
        stop = start + num_dyn
        group_out.append(
            (reord_data[..., start:stop].reshape(base_shape),
             meta_obj.copy(),
             ''))

        # _rf_off
        indices = np.concatenate(
            (np.arange(num_ref), np.arange(-2 * num_ref, -num_ref)))
        group_out.append(
            (reord_data[..., indices].reshape(base_shape),
             meta_obj.copy(),
             '_rf_off'))

        # _rf_grads_ovs_off
        indices = np.concatenate(
            (np.arange(num_ref, 2 * num_ref), np.arange(-num_ref, 0)))
        group_out.append(
            (reord_data[..., indices].reshape(base_shape),
             meta_obj.copy(),
             '_rf_grads_ovs_off'))

    elif (re.search(r'svs_slaser(voi)?_dkd2', seq_file_name) and scan_mode == 2.0):
        num_ref = int(twixObj.hdr['MeasYaps'][('sSpecPara', 'lAutoRefScanNo')])
        num_dyn = int(twixObj.hdr['MeasYaps'][('lAverages',)])

        base_shape = (1, 1, 1) + reord_data.shape[:2] + (-1, )

        for idx, dt in enumerate(dim_tags):
            meta_obj.set_dim_info(idx, dt)

        # Main acq
        start = num_ref
        stop = start + num_dyn
        group_out.append(
            (reord_data[..., start:stop].reshape(base_shape),
             meta_obj.copy(),
             ''))

        # _vapor_ovs_rfoff
        indices = np.concatenate(
            (np.arange(num_ref), np.arange(-num_ref, 0)))
        group_out.append(
            (reord_data[..., indices].reshape(base_shape),
             meta_obj.copy(),
             '_vapor_ovs_rfoff'))

    else:
        print(f'slaser_dkd special case processing: unrecognised case for {seq_file_name} with scan mode {scan_mode}.')
        for idx, dt in enumerate(dim_tags):
            meta_obj.set_dim_info(idx, dt)
        group_out.append(
            (reord_data.reshape((1, 1, 1) + reord_data.shape),
             meta_obj,
             ''))

    return group_out



'''Tests for Philips SDAT/SPAR format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'
list_path = philips_path / 'hyper' / 'raw_226.list'
data_path = philips_path / 'hyper' / 'raw_226.data'
aux_path = philips_path / 'hyper' / 'converted_dcm.dcm'


def test_svs_hyper(tmp_path):

    subprocess.check_call(['spec2nii', 'philips_dl',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           '--special', 'hyper',
                           str(data_path),
                           str(list_path),
                           str(aux_path)])

    img_1 = read_nifti_mrs(tmp_path / 'svs_hyper_short_te.nii.gz')
    img_2 = read_nifti_mrs(tmp_path / 'svs_hyper_edited.nii.gz')
    img_3 = read_nifti_mrs(tmp_path / 'svs_NOI.nii.gz')

    assert img_1.shape == (1, 1, 1, 2048, 15, 32)
    assert np.iscomplexobj(img_1.dataobj)
    assert 1 / img_1.header['pixdim'][4] == 2000.0

    assert img_2.shape == (1, 1, 1, 2048, 15, 4, 56)
    assert np.iscomplexobj(img_2.dataobj)
    assert 1 / img_2.header['pixdim'][4] == 2000.0

    assert img_3.shape == (1, 1, 1, 2048, 15)
    assert np.iscomplexobj(img_2.dataobj)
    assert 1 / img_2.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_1.header.extensions.get_codes()
    hdr_ext = json.loads(img_1.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.74876
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_path.name
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'

    hdr_ext_codes = img_2.header.extensions.get_codes()
    hdr_ext = json.loads(img_2.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_EDIT'
    assert hdr_ext['dim_6_header'] == {'EditCondition': ['A', 'B', 'C', 'D']}
    assert hdr_ext['dim_7'] == 'DIM_DYN'



'''Test special case siemens data.
Includes:
1. Twix SPECIAL (rm_special) data that doesn't report a
    twixObj.hdr.Meas.lFinalMatrixSizeSlice field.
2. Anonymised Siemens DICOM data lacking a number of normal tags.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
data_path_special = siemens_path / 'special' / 'sub-brainphan_nuc-1H_loc-phan_spec_rm-special.dat'
data_path_anon = siemens_path / 'anon' / 'anon_dcm.IMA'


def test_twix_special(tmp_path):

    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'special',
                           '-o', tmp_path,
                           '-t6', 'DIM_USER_0',
                           '-j', str(data_path_special)])

    img_t = read_nifti_mrs(tmp_path / 'special.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 4096, 32, 32)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_USER_0'


def test_dicom_anon(tmp_path):

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'anon',
                           '-o', tmp_path,
                           '-j', str(data_path_anon)])

    img_t = read_nifti_mrs(tmp_path / 'anon.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert img_t.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_t.dataobj)

    assert hdr_ext['SpectrometerFrequency'][0] == 123.255582
    assert hdr_ext['ResonantNucleus'][0] == "1H"
    assert hdr_ext['OriginalFile'][0] == str(data_path_anon)


data_path_slaser_dkd_dicom = siemens_path / 'special_cases_slaser_dkd' / 'DICOM'
slaser_dkd_options = {
    "svs_slaser_dkd_von_wrsoff_13_None": {'': 4, '_rf_off': 0, '_rf_grads_ovs_off': 0},
    "svs_slaser_dkd_von_wrs1_15_None": {'': 4, '_rf_off': 2, '_rf_grads_ovs_off': 2},
    "svs_slaser_dkd_von_wrs2_17_None": {'': 4, '_rf_off': 4, '_rf_grads_ovs_off': 4},
    "svs_slaserVOI_dkd2_von_wrsoff_19_None": {'': 4, '_vapor_ovs_rfoff': 0},
    "svs_slaserVOI_dkd2_von_wrsw1pw3_1_21_None": {'': 4, '_rf_off': 2, '_rf_grads_ovs_off': 2},
    "svs_slaserVOI_dkd2_von_wrsw1pw3_2_23_None": {'': 4, '_rf_off': 4, '_rf_grads_ovs_off': 4},
    "svs_slaserVOI_dkd2_von_wrsw4_1_25_None": {'': 4, '_vapor_ovs_rfoff': 2},
    "svs_slaserVOI_dkd2_von_wrsw4_2_27_None": {'': 4, '_vapor_ovs_rfoff': 4}
}


def test_dicom_slaser_dkd(tmp_path):
    for key in slaser_dkd_options:
        subprocess.run(
            ['spec2nii', 'dicom',
             '-f', key,
             '-o', tmp_path,
             '-j', data_path_slaser_dkd_dicom / key])

        for subkey in slaser_dkd_options[key]:
            if slaser_dkd_options[key][subkey] > 0:
                print(f'{key}, {subkey}, {slaser_dkd_options[key][subkey]}')
                img = read_nifti_mrs(tmp_path / f'{key}{subkey}.nii.gz')
                assert img.shape == (1, 1, 1, 2048, slaser_dkd_options[key][subkey])


data_path_slaser_dkd_twix = siemens_path / 'special_cases_slaser_dkd' / 'twix'
slaser_dkd_options_twix = {
    "meas_MID00053_FID16506_svs_slaser_dkd_von_wrsoff.dat": {'': 4, '_rf_off': 0, '_rf_grads_ovs_off': 0},
    "meas_MID00054_FID16507_svs_slaser_dkd_von_wrs1.dat": {'': 4, '_rf_off': 2, '_rf_grads_ovs_off': 2},
    "meas_MID00055_FID16508_svs_slaser_dkd_von_wrs2.dat": {'': 4, '_rf_off': 4, '_rf_grads_ovs_off': 4},
    "meas_MID00056_FID16509_svs_slaserVOI_dkd2_von_wrsoff.dat": {'': 4, '_vapor_ovs_rfoff': 0},
    "meas_MID00057_FID16510_svs_slaserVOI_dkd2_von_wrsw1pw3_1.dat": {'': 4, '_rf_off': 2, '_rf_grads_ovs_off': 2},
    "meas_MID00058_FID16511_svs_slaserVOI_dkd2_von_wrsw1pw3_2.dat": {'': 4, '_rf_off': 4, '_rf_grads_ovs_off': 4},
    "meas_MID00059_FID16512_svs_slaserVOI_dkd2_von_wrsw4_1.dat": {'': 4, '_vapor_ovs_rfoff': 2},
    "meas_MID00060_FID16513_svs_slaserVOI_dkd2_von_wrsw4_2.dat": {'': 4, '_vapor_ovs_rfoff': 4}
}


def test_twix_slaser_dkd(tmp_path):
    for key in slaser_dkd_options_twix:
        print(key)
        subprocess.run(
            ['spec2nii', 'twix',
             '-e', 'image',
             '-f', key.split('.')[0],
             '-o', tmp_path,
             '-j', data_path_slaser_dkd_twix / key])

        for subkey in slaser_dkd_options_twix[key]:
            if slaser_dkd_options_twix[key][subkey] > 0:
                print(f'{key}, {subkey}, {slaser_dkd_options_twix[key][subkey]}')
                img = read_nifti_mrs(tmp_path / f'{key.split(".")[0]}{subkey}.nii.gz')
                assert img.shape[:4] == (1, 1, 1, 4096)
                assert img.shape[-1] == slaser_dkd_options_twix[key][subkey]



""" Tests for conversion routines for "other" file types.
Currently included:
txt
LCModel Raw

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
"""
import os.path as op
import subprocess
from pathlib import Path

import pytest
import nibabel as nib
import numpy as np


file_path = Path(__file__).parent
testdata = {'txtfile': file_path / 'spec2nii_test_data/other/metab.txt',
            'rawfile': file_path / 'spec2nii_test_data/other/metab.RAW',
            'rawnohdfile': file_path / 'spec2nii_test_data/other/metab_nohdr.RAW',
            'niftifile': file_path / 'spec2nii_test_data/other/metab.nii'}


@pytest.fixture
def affine_file(tmp_path):

    affinepath = op.join(tmp_path, 'affine.txt')

    affine = np.random.random((4, 4))
    affine[3, 0] = 0.0
    affine[3, 1] = 0.0
    affine[3, 2] = 0.0
    affine[3, 3] = 1.0

    np.savetxt(affinepath, affine)

    return affinepath


def test_text(affine_file, tmp_path):
    """Test the 'text' format conversion """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'text',
                     '-f', 'text',
                     '-o', str(tmp_path / "outdir"),
                     '-b', '12000',
                     '-i', '297.219948',
                     '-n', '1H',
                     '-a', affine_file,
                     '-j', testdata['txtfile']])

    # Load the new nifti file
    converted = nib.load(tmp_path / "outdir" / 'text.nii.gz')

    # Compare to original nifti file
    original = nib.load(testdata['niftifile'])

    assert converted.shape == (1, 1, 1, 4096)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)
    assert np.allclose(converted.dataobj[:], original.dataobj[:])
    assert (tmp_path / "outdir" / 'text.json').exists()


def test_raw(affine_file, tmp_path):
    """Test the 'raw' conversion option."""
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'raw',
                     '-f', 'raw',
                     '-n', '1H',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['rawfile']])
    # Load the new nifti file
    converted = nib.load(tmp_path / 'raw.nii.gz')
    # Compare to original nifti file
    original = nib.load(testdata['niftifile'])

    assert converted.shape == (1, 1, 1, 4096)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)
    assert np.allclose(converted.dataobj[:], original.dataobj[:])
    assert (tmp_path / 'raw.json').exists()


def test_raw_nohdr(affine_file, tmp_path):
    """Test the 'raw' conversion + optional arguments."""
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'raw',
                     '-f', 'raw',
                     '-i', '297.219948',
                     '-b', '12000',
                     '-n', '1H',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['rawnohdfile']])
    # Load the new nifti file
    converted = nib.load(tmp_path / 'raw.nii.gz')
    # Compare to original nifti file
    original = nib.load(testdata['niftifile'])

    assert converted.shape == (1, 1, 1, 4096)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)
    assert np.allclose(converted.dataobj[:], original.dataobj[:])
    assert (tmp_path / 'raw.json').exists()



import os.path as op
from pathlib import Path
import subprocess
from PIL import Image, ImageDraw
import pytest

# Data paths
output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

siemens_path = op.join(op.dirname(__file__), 'spec2nii_test_data', 'Siemens')
vb_path = op.join(siemens_path, 'VBData')
ve_path = op.join(siemens_path, 'VEData')

# mrsi VB data
csi_data_names_vb = ['3D_C_S23_5_T20_3_10',
                     '3D_S_T23_5_C20_3_10',
                     '3D_T_C23_5_S20_3_10',
                     'C_S23_5_T20_3_10',
                     'S_T23_5_C20_3_10',
                     'T_C23_5_S20_3_10',
                     'iso_Tra']

csi_Data_dicom_vb = ['DICOM/csi_se_3D_C>S23.5>T20.3_10_8_1',
                     'DICOM/csi_se_3D_S>T23.5>C20.3_10_7_1',
                     'DICOM/csi_se_3D_T>C23.5>S20.3_10_6_1',
                     'DICOM/csi_se_C>S23.5>T20.3_10_5_1',
                     'DICOM/csi_se_S>T23.5>C20.3_10_4_1',
                     'DICOM/csi_se_T>C23.5>S20.3_10_3_1',
                     'DICOM/csi_se_Tra_sat_9_1']

screen_shots_vb = ['Screenshots/csi_se_3D_C_S23_5_T20_3_10.png',
                   'Screenshots/csi_se_3D_S_T23_5_C20_3_10.png',
                   'Screenshots/csi_se_3D_T_C23_5_S20_3_10.png',
                   'Screenshots/csi_se_C_S23_5_T20_3_10.png',
                   'Screenshots/csi_se_S_T23_5_C20_3_10.png',
                   'Screenshots/csi_se_T_C23_5_S20_3_10.png',
                   'Screenshots/csi_se_Tra_sat.png']

# MRSI VE data
csi_data_names_ve = ['3D_c_s23_5_t20_3_R10',
                     '3D_s_t23_5_c20_3_R10',
                     '3D_t_c23_5_s20_3_R10',
                     'c_s23_5_t20_3_R10',
                     'iso_tra',
                     's_t23_5_c20_3_R10',
                     't_c23_5_s20_3_R10']

csi_Data_dicom_ve = ['DICOM/csi_se_3D_c>s23.5>t20.3_R10_9_1',
                     'DICOM/csi_se_3D_s>t23.5>c20.3_R10_8_1',
                     'DICOM/csi_se_3D_t>c23.5>s20.3_R10_7_1',
                     'DICOM/csi_se_c>s23.5>t20.3_R10_6_1',
                     'DICOM/csi_se_iso_tra_sat_14_1',
                     'DICOM/csi_se_s>t23.5>c20.3_R10_5_1',
                     'DICOM/csi_se_t>c23.5>s20.3_R10_4_1']

screen_shots_ve = ['Screenshots/csi_se_3D_c_s23_5_t20_3_R10.png',
                   'Screenshots/csi_se_3D_s_t23_5_c20_3_R10.png',
                   'Screenshots/csi_se_3D_t_c23_5_s20_3_R10.png',
                   'Screenshots/csi_se_c_s23_5_t20_3_R10.png',
                   'Screenshots/csi_se_iso_tra_sat.png',
                   'Screenshots/csi_se_s_t23_5_c20_3_R10.png',
                   'Screenshots/csi_se_t_c23_5_s20_3_R10.png']


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def crop_and_flip_first_third(img_in):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 180, int(width / 3), height - 180))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 1, 180, width, height - 180))
    out = Image.new('RGB', (width, height - 360))
    out.paste(cropped_0, (0, 0))
    out.paste(cropped_1, (int(width / 3) + 1, 0))
    return out


# Test siemens VB svs
@pytest.mark.orientation
def test_VB(tmp_path):
    sub_images = []
    for idx, (f_d, name) in enumerate(zip(csi_Data_dicom_vb, csi_data_names_vb)):
        # Convert DICOM
        subprocess.check_call(['spec2nii', 'dicom',
                               '-f', name + '_d',
                               '-o', tmp_path,
                               '-j', op.join(vb_path, f_d)])

        # Make fsleyes rendering
        if name == 'iso_Tra':
            x, y, z = '95', '89', '99'
        else:
            x, y, z = '95', '88', '95'

        subprocess.check_call(['fsleyes',
                               'render', '-of', op.join(tmp_path, f'csi_{idx}.png'),
                               '-vl', x, y, z,
                               '-hc', op.join(vb_path, 'T1.nii.gz'),
                               op.join(tmp_path, name + '_d.nii.gz'), '-ot', 'complex',
                               '-a', '60', '-cm', 'red'])

        fsl_ss = Image.open(op.join(tmp_path, f'csi_{idx}.png'))
        width, height = fsl_ss.size
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = Image.open(op.join(vb_path, screen_shots_vb[idx]))
        width, height = ss.size
        ss_cropped = ss.crop((0, 50, width - 50, height / 2 - 80))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 4))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * c, si.height * r))
        draw.text((10 + si.width * c, 10 + si.height * r),
                  csi_data_names_vb[idx],
                  (255, 0, 0))

    final_img.save(output_path / 'csi_vb.png')


@pytest.mark.orientation
def test_VE(tmp_path):
    sub_images = []
    for idx, (f_d, name) in enumerate(zip(csi_Data_dicom_ve, csi_data_names_ve)):

        # Convert DICOM
        subprocess.check_call(['spec2nii', 'dicom',
                               '-f', name + '_d',
                               '-o', tmp_path,
                               '-j', op.join(ve_path, f_d)])

        # Make fsleyes rendering
        x, y, z = '57', '63', '37'

        subprocess.check_call(['fsleyes',
                               'render', '-of', op.join(tmp_path, f'csi_{idx}.png'),
                               '-vl', x, y, z,
                               '-hc', op.join(ve_path, 'T1.nii.gz'),
                               op.join(tmp_path, name + '_d.nii.gz'), '-ot', 'complex',
                               '-a', '60', '-cm', 'red'])

        fsl_ss = Image.open(op.join(tmp_path, f'csi_{idx}.png'))
        width, height = fsl_ss.size
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = Image.open(op.join(ve_path, screen_shots_ve[idx]))
        width, height = ss.size
        ss_cropped = ss.crop((0, 50, width - 50, height / 2 - 80))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 4))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * c, si.height * r))
        draw.text((10 + si.width * c, 10 + si.height * r),
                  csi_data_names_ve[idx],
                  (255, 0, 0))

    final_img.save(output_path / 'csi_ve.png')



"""spec2nii module containing functions specific to interpreting various formats
Contains text and LCModel formats.
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
from datetime import datetime
from os.path import basename, splitext

import numpy as np
from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii import __version__ as spec2nii_ver
from spec2nii.nifti_orientation import NIFTIOrient


class InsufficentHeaderInformationError(Exception):
    pass


def text(args):
    '''Processing for simple ascii formatted columns of data.'''
    # Read text from file
    data = np.loadtxt(args.file)
    data = data[:, 0] + 1j * data[:, 1]

    newshape = (1, 1, 1) + data.shape
    data = data.reshape(newshape)

    # Interpret required arguments (frequency and bandwidth)
    dwelltime = 1.0 / args.bandwidth

    meta = Hdr_Ext(
        args.imagingfreq,
        args.nucleus)

    meta.set_standard_def('ConversionMethod', 'spec2nii')
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)
    meta.set_standard_def('OriginalFile', [basename(args.file), ])

    # Read optional affine file
    if args.affine:
        affine = np.loadtxt(args.affine)
    else:
        tmp = np.array([10000, 10000, 10000, 1])
        affine = np.diag(tmp)

    nifti_orientation = NIFTIOrient(affine)

    img_out = [gen_nifti_mrs_hdr_ext(
        data,
        dwelltime,
        meta,
        nifti_orientation.Q44,
        no_conj=True), ]

    # File names
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [splitext(basename(args.file))[0], ]

    # Place in data output format
    return img_out, fname_out


def lcm_raw(args):
    '''Processing for LCModel .RAW (and .H2O) files.
    Currently only handles one FID per file.
    '''
    # Read data from file
    data, header = readLCModelRaw(args.file, conjugate=True)

    newshape = (1, 1, 1) + data.shape
    data = data.reshape(newshape)

    # meta
    try:
        dwelltime = header['dwelltime']
    except KeyError:
        if args.bandwidth is not None:
            dwelltime = 1 / args.bandwidth
        else:
            raise InsufficentHeaderInformationError(
                "There is no 'dwelltime' key in the header information, "
                "please use the optional '-b' argument to provide a bandwidth.")

    spec_frequency = header['centralFrequency']
    if spec_frequency is None:
        if args.imagingfreq is not None:
            spec_frequency = args.imagingfreq
        else:
            raise InsufficentHeaderInformationError(
                "There is no 'centralFrequency' key in the header information, "
                "please use the optional '-i' argument to provide a frequency.")

    meta = Hdr_Ext(
        spec_frequency,
        args.nucleus)

    meta.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)
    meta.set_standard_def('OriginalFile', [basename(args.file), ])

    # Read optional affine file
    if args.affine:
        affine = np.loadtxt(args.affine)
    else:
        tmp = np.array([10000, 10000, 10000, 1])
        affine = np.diag(tmp)

    nifti_orientation = NIFTIOrient(affine)

    img_out = [gen_nifti_mrs_hdr_ext(
        data,
        dwelltime,
        meta,
        nifti_orientation.Q44,
        no_conj=True), ]

    # File names
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [splitext(basename(args.file))[0], ]

    # Place in data output format
    return img_out, fname_out


def readLCModelRaw(filename, conjugate=True):
    """
    Read .RAW (or.H2O) format file
    Parameters
    ----------
    filename : string
        Name of .RAW file

    Returns
    -------
    array-like
        Complex data
    dict
        Header information

    """
    header = []
    data   = []
    in_header = False
    with open(filename, 'r') as f:
        for line in f:
            if (line.find('$') > 0):
                in_header = True
            if in_header:
                header.append(line)
            else:
                data.append(list(map(float, line.split())))

            if line.find('$END') > 0:
                in_header = False

    # Reshape data
    data = np.concatenate([np.array(i) for i in data])
    data = (data[0::2] + 1j * data[1::2]).astype(complex)

    # LCModel-specific conjugation
    if conjugate:
        data = np.conj(data)

    # Tidy header info
    header = unpackHeader(header)

    return data, header


def unpackHeader(header):
    """
       Extracts useful info from header into dict

       Including central frequency, dwelltime, echotime
    """

    def tidy(x):
        return x.lower().replace(',', '')

    tidy_header = dict()
    tidy_header['centralFrequency'] = None
    tidy_header['bandwidth'] = None
    tidy_header['echotime'] = None
    for line in header:
        if line.lower().find('hzpppm') > 0:
            tidy_header['centralFrequency'] = float(tidy(line).split()[-1]) * 1E6
        if line.lower().find('dwelltime') > 0:
            tidy_header['dwelltime'] = float(tidy(line).split()[-1])
            tidy_header['bandwidth'] = 1 / float(tidy(line).split()[-1])
        if line.lower().find('deltat') > 0:
            tidy_header['dwelltime'] = float(tidy(line).split()[-1])
            tidy_header['bandwidth'] = 1 / float(tidy(line).split()[-1])
        if line.lower().find('echot') > 0:
            tidy_header['echotime'] = float(tidy(line).split()[-1]) / 1e3
        if line.lower().find('badelt') > 0:
            tidy_header['dwelltime'] = float(tidy(line).split()[-1])
            tidy_header['bandwidth'] = 1 / float(tidy(line).split()[-1])

    return tidy_header



'''Test the ability to choose between nifti 1 and nifti2 file types.'''

import subprocess
import nibabel as nib
from pathlib import Path
import numpy as np
from nifti_mrs.nifti_mrs import NIFTI_MRS

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
vb_path = siemens_path / 'VBData'
data_path = vb_path / 'DICOM/svs_se_C>T15>S10_10_12_1'


def test_process(tmp_path):
    # Convert DICOM
    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'nifti_2',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    subprocess.check_call(['spec2nii', 'dicom',
                           '-f', 'nifti_1',
                           '--nifti1',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    img_nifti2 = nib.load(tmp_path / 'nifti_2.nii.gz')
    img_nifti1 = nib.load(tmp_path / 'nifti_1.nii.gz')

    assert type(img_nifti2) is nib.nifti2.Nifti2Image
    assert type(img_nifti1) is nib.nifti1.Nifti1Image

    img_nifti2 = NIFTI_MRS(tmp_path / 'nifti_2.nii.gz')
    img_nifti1 = NIFTI_MRS(tmp_path / 'nifti_1.nii.gz')

    assert type(img_nifti2.header) is nib.nifti2.Nifti2Header
    assert type(img_nifti1.header) is nib.nifti1.Nifti1Header

    # Check parameters stored are the same
    assert np.allclose(img_nifti2[:], img_nifti1[:])
    assert np.isclose(img_nifti2.dwelltime, img_nifti1.dwelltime)
    assert img_nifti2.hdr_ext.current_keys == img_nifti1._hdr_ext.current_keys



"""
fileiobase provides general purpose NMR file IO functions and classes
used by multiple nmrglue.fileio modules.

Originally part of NMR glue -- an open source module for working with NMR data
in Python https://github.com/jjhelmus/nmrglue/

If you find nmrglue useful in your research please cite the package as:

J.J. Helmus, C.P. Jaroniec, Nmrglue: An open source Python package for the
analysis of multidimensional NMR data, J. Biomol. NMR 2013, 55, 355-367,
https://doi.org/10.1007/s10858-013-9718-x.
----------------

Copyright Notice and Statement for the nmrglue Project


Copyright (c) 2010-2015 Jonathan J. Helmus
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:


a. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.


b. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.


c. Neither the name of the author nor the names of contributors may
   be used to endorse or promote products derived from this software
   without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

---------------

"""


import os
import string
import itertools
from functools import reduce

import numpy as np


def create_blank_udic(ndim):
    """
    Create a blank universal dictionary for a spectrum of dimension ndim.
    """
    udic = {
        "ndim": ndim,
    }

    for i in range(ndim):
        udic[i] = {
            "sw": 999.99,       # spectral width in Hz
            "complex": True,    # Quadrature, True when dimension is complex
            "obs": 999.99,      # Observation frequency in MHz
            "car": 999.99,      # Carrier frequency in Hz
            # Number of points in dimension based on the shape of the data array.
            # As such the direct dimension (-1) size is R|I, all indirect
            # dimensions are R+I
            "size": 1,
            "label": ["X", "Y", "Z", "A"][i],   # name of dimension
            # encoding of dimension, ie states, tppi, etc.  The direct dimension
            # should be listed as direct.
            "encoding": "direct" if i == ndim - 1 else "states",
            # time and freq flags for domain of dimension
            "time": True,
            "freq": False,
        }

    return udic


class unit_conversion():
    """
    Provides methods to convert between common NMR units

    Parameters
    ----------
    size : int
        Number of points in dimension (R|I).
    cplex : bool
        True if dimension is complex, False is real.
    sw : float
        Spectral width in Hz.
    obs : float
        Observation frequency in MHz.
    car : float
        Carrier frequency in Hz.

    """
    def __init__(self, size, cplx, sw, obs, car):
        """
        create and set up a unit_conversion object
        """
        # fundamental units
        self._size = int(size)
        self._cplx = bool(cplx)
        self._sw = float(sw)
        self._obs = float(obs)
        self._car = float(car)

        # derived units (these are in ppm)
        self._delta = -self._sw / (self._size * self._obs)
        self._first = self._car / self._obs - self._delta * self._size / 2.

    # individual unit conversion functions
    def __percent2pts(self, percent):
        return percent * (self._size - 1) / 100.0

    def __pts2percent(self, pts):
        return pts * 100 / (self._size - 1.0)

    def __hz2pts(self, hz):
        return ((hz / self._obs) - self._first) / self._delta

    def __pts2hz(self, pts):
        return (pts * self._delta + self._first) * self._obs

    def __ppm2pts(self, ppm):
        return (ppm - self._first) / self._delta

    def __pts2ppm(self, pts):
        return (pts * self._delta) + self._first

    # times based units: seconds, ms, and us
    def __sec2pts(self, sec):
        return sec * self._sw

    def __pts2sec(self, pts):
        return pts * 1. / self._sw

    def __ms2pts(self, ms):
        return ms * self._sw / 1.e3

    def __pts2ms(self, pts):
        return pts * 1.e3 / self._sw

    def __us2pts(self, us):
        return us * self._sw / 1.e6

    def __pts2us(self, pts):
        return pts * 1.e6 / self._sw

    # routers
    def __unit2pnt(self, val, units):
        """
        Convert units to points
        """
        units = units.upper()
        if units == "PPM":
            pts = self.__ppm2pts(val)
        elif units == "HZ":
            pts = self.__hz2pts(val)
        elif units in ("%", "PERCENT"):
            pts = self.__percent2pts(val)
        elif units in ("SEC", "SECOND", "S"):
            pts = self.__sec2pts(val)
        elif units == "MS":
            pts = self.__ms2pts(val)
        elif units == "US":
            pts = self.__us2pts(val)
        else:
            raise ValueError("invalid unit type")
        # if self._cplx:
        #     return pts+round(pts)
        # else:
        return pts

    def __pnt2unit(self, val, units):
        """
        Convert points to units
        """
        units = units.upper()
        # if self._cplx:
        #     val = val-round(val)
        if units == "PPM":
            k = self.__pts2ppm(val)
        elif units == "HZ":
            k = self.__pts2hz(val)
        elif units in ("%", "PERCENT"):
            k = self.__pts2percent(val)
        elif units in ("SEC", "SECOND", "S"):
            k = self.__pts2sec(val)
        elif units == "MS":
            k = self.__pts2ms(val)
        elif units == "US":
            k = self.__pts2us(val)
        else:
            raise ValueError("invalid units")
        return k

    def __str2pnt(self, s):
        """
        Convert string with units to points
        """
        units = s.strip(string.digits + string.whitespace + "." + "-").upper()
        val = float(s.strip(string.ascii_letters + string.whitespace + "%"))
        return self.__unit2pnt(val, units)

    def __convert(self, val, unit=None):
        """
        Convert string or value/unit pair
        """
        if isinstance(val, str):
            return self.__str2pnt(val)
        else:
            if unit is None:
                raise ValueError("invalid unit type")
            return self.__unit2pnt(val, unit)

    # User functions
    def f(self, val, unit=None):
        """
        Convert string or value/unit pair to float
        """
        return self.__convert(val, unit)

    def i(self, val, unit=None):
        """
        Convert string or value/unit pair to integer
        """
        return int(round(self.__convert(val, unit)))

    def ppm(self, val):
        """
        Convert to ppm
        """
        return self.__pnt2unit(val, "PPM")

    def hz(self, val):
        """
        Convert to Hz
        """
        return self.__pnt2unit(val, "HZ")

    def percent(self, val):
        """
        Convert to percent
        """
        return self.__pnt2unit(val, "PERCENT")

    def seconds(self, val):
        """
        Convert to seconds
        """
        return self.__pnt2unit(val, "SEC")

    def sec(self, val):
        """
        Convert to seconds
        """
        return self.__pnt2unit(val, "SEC")

    def ms(self, val):
        """
        Convert to milliseconds (ms)
        """
        return self.__pnt2unit(val, "MS")

    def us(self, val):
        """
        Convert to microseconds (us)
        """
        return self.__pnt2unit(val, "US")

    def unit(self, val, unit):
        """
        Convert val points to unit
        """
        return self.__pnt2unit(val, unit)

    # limits and scales
    @staticmethod
    def percent_limits():
        """
        Return tuple of left and right edges in percent
        """
        return 0.0, 100.0

    def percent_scale(self):
        """
        Return array of percent values
        """
        return np.linspace(0.0, 100.0, self._size)

    def ppm_limits(self):
        """
        Return tuple of left and right edges in ppm
        """
        return self.ppm(0), self.ppm(self._size - 1)

    def ppm_scale(self):
        """
        Return array of ppm values
        """
        x0, x1 = self.ppm_limits()
        return np.linspace(x0, x1, self._size)

    def hz_limits(self):
        """
        Return tuple of left and right edges in Hz
        """
        return self.hz(0), self.hz(self._size - 1)

    def hz_scale(self):
        """
        Return array of Hz values
        """
        x0, x1 = self.hz_limits()
        return np.linspace(x0, x1, self._size)

    def sec_limits(self):
        """
        Return tuple of left and right edges in seconds
        """
        return self.sec(0), self.sec(self._size - 1)

    def sec_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.sec_limits()
        return np.linspace(x0, x1, self._size)

    def ms_limits(self):
        """
        Return tuple of left and right edges in milliseconds
        """
        return self.ms(0), self.ms(self._size - 1)

    def ms_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.ms_limits()
        return np.linspace(x0, x1, self._size)

    def us_limits(self):
        """
        Return tuple of left and right edges in milliseconds
        """
        return self.us(0), self.us(self._size - 1)

    def us_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.us_limits()
        return np.linspace(x0, x1, self._size)

    __call__ = i    # calling the object x is the same as x.i


def uc_from_udic(udic, dim=-1):
    """
    Create a unit conversion object from a Universal dictionary.

    Parameters
    ----------
    udic : dic
        Universal dictionary of spectral parameters.
    dim : int. optional
        Dimension number to create unit conversion object for.  Default is for
        last dimension.

    Returns
    -------
    uc : unit conversion object.
        Unit conversion object for given dimension.
    """
    if dim == -1:
        dim = udic['ndim'] - 1  # last dimension
    adic = udic[dim]
    return unit_conversion(adic['size'], adic['complex'], adic['sw'],
                           adic['obs'], adic['car'])


def uc_from_freqscale(scale, obs, unit='ppm'):
    """
    Create a unit conversion object from a spectrum frequency scale axis.

    Parameters
    ----------
    scale : array like
        array of spectrum axis
    obs : float
        Observation frequency in MHz.
    unit: {'ppm', 'hz', 'khz'}
        The unit of the scale axis.

    Returns
    -------
    uc : unit conversion object.
        Unit conversion object for given axis.
    """
    scale = np.array(scale)
    size = len(scale)

    if unit in ('ppm', 'hz', 'khz'):
        cplex = False

        min = scale.min()
        max = scale.max()

        # The scale needs be corrected by extending each extremum by half the
        # bin width (to convert from centers to edges).
        dx = abs(scale[1] - scale[0])

        if unit == 'ppm':
            sw = ((max + dx / 2.0) - (min - dx / 2.0)) * obs
            car = (min - dx / 2.0 + (max - min) / 2.0) * obs
        elif unit == 'hz':
            sw = ((max + dx / 2.0) - (min - dx / 2.0))
            car = (min - dx / 2.0 + (max - min) / 2.0)
        else:
            # unit is 'kHz':
            sw = ((max + dx / 2.0) - (min - dx / 2.0)) / 1.e3
            car = (min - dx / 2.0 + (max - min) / 2.0) / 1.e3

    else:
        mesg = f'{unit} is not a supported unit.'
        raise ValueError(mesg)

    return unit_conversion(size, cplex, sw, obs, car)


def open_towrite(filename, overwrite=False, mode='wb'):
    """
    Open filename for writing and return file object

    Function checks if file exists (and raises IOError if overwrite=False) and
    creates necessary directories as needed.
    """
    # check if file exists and overwrite if False
    if os.path.exists(filename) and (overwrite is False):
        raise OSError("File exists, recall with overwrite=True")

    p, fn = os.path.split(filename)  # split into filename and path
    # create directories if needed
    if p != '' and os.path.exists(p) is False:
        os.makedirs(p)

    return open(filename, mode)

################################################
# numpy ndarray emulation and helper functions #
################################################

# iterators for ND array


def ndfrom_iter(shape, slices):
    ch = [range(lenx)[sX] for lenx, sX in zip(shape, slices)]
    return itertools.product(*ch)


def ndto_iter(shape, slices):
    ich = [range(len(range(lenx)[sX])) for lenx, sX in zip(shape, slices)]
    return itertools.product(*ich)


def ndtofrom_iter(shape, slices):
    ch = [range(lenx)[sX] for lenx, sX in zip(shape, slices)]
    ich = [range(len(i)) for i in ch]
    return zip(itertools.product(*ich), itertools.product(*ch))


def size_and_ndtofrom_iter(shape, slices):
    ch = [range(lenx)[sX] for lenx, sX in zip(shape, slices)]
    s = [len(i) for i in ch]
    ich = [range(i) for i in s]
    return s, zip(itertools.product(*ich), itertools.product(*ch))


# index2trace and trace2index functions


def index2trace_flat(shape, index):
    """
    Calculate trace number from shape and index of all indirect dimensions
    assuming a flat structure
    """
    # We need to perform:
    # index[0]*shape[1]*...shape[-1] + index[1]*shape[2]*...shape[-1] + ...
    # + index[-1]*shape[-1] + index[-1]
    # To do this we calculate the product of shape[X] elements and multiple
    # by the corresponding index element, index[-1] as added at the beginning
    a = index[-1]
    for i, v in enumerate(index[:-1]):
        mult = reduce(lambda x, y: x * y, shape[i + 1:])
        a = a + mult * v
    return a


def trace2index_flat(shape, ntrace):
    """
    Calculate the index of a trace assuming a flat structure
    """
    # algorithm is to take quotient/remainders of sizes in reverse
    q = ntrace  # seed quotient with remained
    index = []
    for s in shape[:0:-1]:  # loop from last size to 2nd size
        q, r = divmod(q, s)
        index.insert(0, r)
    index.insert(0, q)
    return tuple(index)


def index2trace_opp(shape, index):
    """
    Calculate trace number from shape and index of all indirect dimensions
    assuming a phase ordering opposite the time increments.
    """
    n = len(shape)
    # deal with the phase component
    phases = [v % 2 for v in index]
    nphase = index2trace_flat([2] * n, phases[::-1])
    # deal with the remainder
    pindex = [v // 2 for v in index]
    pshape = [i // 2 for i in shape]
    nbase = index2trace_flat(pshape, pindex)
    return nbase * 2 ** n + nphase


def trace2index_opp(shape, ntrace):
    """
    Calculate the index of a trace assuming opposite phase/time increment
    ordering
    """
    n = len(shape)
    q, r = divmod(ntrace, 2 ** n)
    to_add = list(trace2index_flat([2] * n, r))[::-1]
    pshape = [i // 2 for i in shape]
    base = list(trace2index_flat(pshape, q))
    total = [b * 2 + a for b, a in zip(base, to_add)]
    return tuple(total)


def index2trace_reg(shape, index):
    """
    Calculate trace number from shape and index of all indirect dimensions
    assuming the same  phase and time ordering.
    """
    n = len(shape)
    # deal with the phase component
    phases = [v % 2 for v in index]
    nphase = index2trace_flat([2] * n, phases)
    # deal with the remainder
    pindex = [v // 2 for v in index]
    pshape = [i // 2 for i in shape]
    nbase = index2trace_flat(pshape, pindex)
    return nbase * 2 ** n + nphase


def trace2index_reg(shape, ntrace):
    """
    Calculate the index of a trace assuming the same phase/time increment
    ordering
    """
    n = len(shape)
    q, r = divmod(ntrace, 2 ** n)
    to_add = list(trace2index_flat([2] * n, r))
    pshape = [i // 2 for i in shape]
    base = list(trace2index_flat(pshape, q))
    total = [b * 2 + a for b, a in zip(base, to_add)]
    return tuple(total)

#
# data_nd class
#
# inherited classes should define:
#
#    __init__ which sets up the object and defines at minimum
#
#       self.fshape shape of data on disk (shape when order = (0,1,2...)
#       self.order order of axes, default is (0,1,2,...)
#       self.dtype
#
#   self.__setdimandshape__ can be called to set self.dim and self.shape
#    if they are not set by __init__
#
#    __fgetitem__ which takes well formatted tuples of slices
#    and returns ndarray objects
#
#    __fcopy__ which creates a copy provided only self and order parameters
#


class data_nd:
    """
    Base class for building objects which emulate ndarray objects without
    loading data into memory.  These object have the following properties:

    * slicing operations return ndarray objects
    * can iterate over with expected results
    * transpose and swapaxes functions create a new data_nd object with the
      new axes ordering
    * has ndim, shape, and dtype attributes.

    Notes
    -----

    Classes which are use this class as a base should define the following
    methods:

    __init__ which must set up the object and defines at minimum:

        self.fshape : tuple
            Shape of the data on disk, the shape when order = (0, 1, 2, ..)
        self.order : tuple
            Ordering of the axes
        self.dtype : dtype
            Dtype of the emulated ndarray

        self.__setdimandshape__ should be called if self.dim and self.shape
            are not set up __init__.

    __fgetitem__ which takes a well formatted tuple of slices and returns a
        ndarray object with the selected data

    __fcopy__(self, order) which created a copy of the object with the new
        order

    """

    def __init__(self, order):
        pass

    def __setdimandshape__(self):
        """ Set the ndim and shape attributes from fshape """
        # set ndim and shape
        self.ndim = len(self.fshape)
        self.shape = tuple(self.fshape[i] for i in self.order)

    def __copy__(self):
        """
        create a copy
        """
        return self.__fcopy__(self, self.order)

    def __getitem__(self, key):
        """
        x.__getitem__(y) <==> x[y]
        """
        # convert the key into a list
        if not isinstance(key, tuple):
            rlist = [key]
        else:
            rlist = list(key)

        # remove Ellipsis
        while Ellipsis in rlist:
            i = rlist.index(Ellipsis)
            rlist.pop(i)
            for j in range(self.ndim - len(rlist)):
                rlist.insert(i, slice(None))

        if len(rlist) > self.ndim:
            raise IndexError("invalid index")

        # replace none-slices with slice objects
        for i, v in enumerate(rlist):
            if not isinstance(v, slice):

                # check for out of range indexes
                if v >= self.shape[i]:
                    raise IndexError("index({}) out of range(0 <= index < {}) \
                    in dimension {}".format(v, self.shape[i] - 1, i))

                if v <= (-1 * self.shape[i] - 1):
                    raise IndexError("index({}) out of range(0 <= index < {}) \
                    in dimension {}".format(v, self.shape[i] - 1, i))

                if v < 0:
                    w = self.shape[i] + v
                    rlist[i] = slice(w, w + 1, 1)
                else:
                    rlist[i] = slice(v, v + 1, 1)

        # pad the list with additional dimentions
        for i in range(len(rlist), self.ndim):
            rlist.append(slice(None))

        # reorder the slices into file order
        frlist = [rlist[self.order.index(i)] for i in range(self.ndim)]

        # get the data from disk
        data = self.__fgetitem__(tuple(frlist))

        # re-order the data
        if data.shape != (0,):
            return np.squeeze(data.transpose(self.order))
        else:
            return data

    def __len__(self):
        """
        x._len__ <==> len(x)
        """
        return self.shape[0]

    def __iter__(self):
        """ x.__iter__() <==> iter(x) """
        for index in range(0, self.shape[0]):
            yield self[index]

    def swapaxes(self, axis1, axis2):
        """
        Return object with `axis1` and `axis2` interchanged.
        """
        axis1, axis2 = int(axis1), int(axis2)
        if axis1 < 0:
            axis1 = self.ndim - axis1
        if axis2 < 0:
            axis2 = self.ndim - axis2
        if axis1 >= self.ndim:
            raise ValueError("bad axis1 argument to swapaxes")
        if axis2 >= self.ndim:
            raise ValueError("bad axis2 argument to swapaxes")

        order = list(self.order)
        order[axis1], order[axis2] = order[axis2], order[axis1]
        n = self.__fcopy__(order=order)
        return n

    def transpose(self, *axes):
        """
        Return object with axes transposed.

        Parameters
        ----------
        axes : None, tuple or ints, or `n` ints
            * None or no arguments: reverse order of the axes

            * tuple of ints: `i` in the `j`-th place in the tuples means the
            'i'-th axis becomes the new objects `j`-th axis.

            * `n` ints: same as an n-tuple.

        Returns
        -------
        out : data_nd object
            Object whose axes are permuted.

        """
        if axes == ():    # default is to switch order of axes
            axes = range(self.ndim)[::-1]

        if len(axes) == 1:    # if a single tuple is given unpack
            axes = axes[0]

        try:    # convert to integers
            axes = [int(i) for i in axes]
        except Exception:
            raise TypeError("an integer is required")

        if len(axes) != self.ndim:   # check for to few/many axes
            raise ValueError("axes don't match array")

        # replace negatives axes values with positives
        for i, v in enumerate(axes):
            if v < 0:
                axes[i] = self.ndim + v

        # check for invalid axes
        for v in axes:
            if v >= self.ndim:
                raise ValueError("invalid axis for this array")

        # check for repeated axes
        if len(set(axes)) != self.ndim:
            raise ValueError("repeated axis in tranpose")

        # create a new data_nd object with transposed order
        return self.__fcopy__(order=tuple(self.order[i] for i in axes))



'''Orientation tests for GE pfile format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path

import pytest
from PIL import Image, ImageDraw

ge_path = Path(__file__).parent / 'spec2nii_test_data' / 'ge'

output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

# SVS Data paths
'''1.  Filename:  P03072.7
Voxel position:  RL 0 - AP 0 - SI 0
Voxel size:  x 20 - y 25 z 30
Rotation: None

2.  Filename:  P06656.7
Voxel position:  R 10 - P 20 - S 30
Voxel size:  x 20 - y 25 z 30
Rotation: Ax 10 deg cw - Cor 30 deg cw

3.  Filename:  P08704.7
Voxel position:  L 20 - P 40 - I 60
Voxel size:  x 20 - y 25 z 30
Rotation: Cor 10 deg cw - Sag 30 deg cw

4.  Filename:  P10240.7
Voxel position:  R 30 - P 60 - S 10
Voxel size:  x 20 - y 25 z 30
Rotation: Sag 10 deg ccw - Ax 30 deg ccw
'''

svs_path = [ge_path / 'pFiles' / 'svMRS' / 'P03072.7',
            ge_path / 'pFiles' / 'svMRS' / 'P06656.7',
            ge_path / 'pFiles' / 'svMRS' / 'P08704.7',
            ge_path / 'pFiles' / 'svMRS' / 'P10240.7']

ss_path = ge_path / 'screenshots'

t1_pos = [(98, 158, 149),
          (112, 138, 180),
          (79, 117, 90),
          (125, 97, 158)]

# MRSI Data paths
'''1.  Filename:  P13824.7
Start:  S 4.8 - R 41.3 - A 3.0
End:  I 5.1 - L 44.7 - P 66.3
Rotation:  None

2.  Filename: P15360.7
Start:  S 37.0 - R 41.7 - P 0.4
End:  S 27.0 - L 44.3 - P 69.8
Rotation: Cor 10 deg cw - Ax 30 deg cw

3.  Filename: P16896.7
Start:  S 5.0 - L 5.8 - A 44.5
End:  I 4.9 - L 91.9 - P 24.9
Rotation: Sag 10 deg cw - Cor 30 deg cw

4.  Filename: P18432.7
Start:  S 6.0 - R 79.4 - P 31.1
End:  I 3.9 - L 6.6 - P 100.6
Rotation: Ax 10 deg ccw - Sag 30 deg ccw
'''

mrsi_path = [ge_path / 'pFiles' / 'MRSI' / 'P13824.7',
             ge_path / 'pFiles' / 'MRSI' / 'P15360.7',
             ge_path / 'pFiles' / 'MRSI' / 'P16896.7',
             ge_path / 'pFiles' / 'MRSI' / 'P18432.7']

mrsi_dcm_path = [ge_path / 'from_dicom' / '9_PROBE_3D_CSI-no_rotation_cNova32ch_i00001.nii.gz',
                 ge_path / 'from_dicom' / '10_PROBE_3D_CSI-rotation_1_cNova32ch_i00001.nii.gz',
                 ge_path / 'from_dicom' / '11_PROBE_3D_CSI-rotation_2_cNova32ch_i00001.nii.gz',
                 ge_path / 'from_dicom' / '12_PROBE_3D_CSI-rotation_3_cNova32ch_i00001.nii.gz']

mrsi_t1_pos = [(98, 128, 149),
               (98, 118, 180),
               (72, 166, 157),
               (130, 102, 146)]


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def get_concat_h(im1, im2, im3):
    dst = Image.new('RGB',
                    (im1.width + im2.width + im3.width,
                     max([im1.height, im2.height, im3.height])))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (im1.width, 0))
    dst.paste(im3, (im1.width + im2.width, 0))
    return dst


def crop_and_flip_first_third(img_in):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 180, int(width / 3), height - 180))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 1, 180, width, height - 180))
    out = Image.new('RGB', (width, height - 360))
    out.paste(cropped_0, (0, 0))
    out.paste(cropped_1, (int(width / 3) + 1, 0))
    return out


@pytest.mark.orientation
def test_svs_orientation(tmp_path):

    sub_images = []
    for idx, (pfile, pos) in enumerate(zip(svs_path, t1_pos)):
        subprocess.check_call(['spec2nii', 'ge',
                               '-f', 'svs',
                               '-o', tmp_path,
                               '-j',
                               str(pfile)])

        # Make fsleyes rendering
        subprocess.check_call(['fsleyes', 'render',
                               '-of', tmp_path / f'svs_{idx}.png',
                               '-vl', str(pos[0]), str(pos[1]), str(pos[2]),
                               '-xc', '0', '0', '-yc', '0', '0', '-zc', '0', '0',
                               '-hc', ge_path / 'from_dicom' / 'T1.nii.gz',
                               '-dr', '-211', '7400',
                               tmp_path / 'svs.nii.gz', '-ot', 'complex',
                               '-a', '50', '-cm', 'blue'])

        fsl_ss = Image.open(tmp_path / f'svs_{idx}.png')
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = get_concat_h(Image.open(ss_path / f'svs_{idx + 1}.0001.jpg'),
                          Image.open(ss_path / f'svs_{idx + 1}.0002.jpg'),
                          Image.open(ss_path / f'svs_{idx + 1}.0003.jpg'))
        ss_cropped = ss  # ss.crop((0, 110, width - 50, height / 2 - 30))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 2))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * r, si.height * c))
        draw.text((10 + si.width * r, 10 + si.height * c), svs_path[idx].stem, (255, 0, 0))

    final_img.save(output_path / 'ge_svs.png')


@pytest.mark.orientation
def test_mrsi_orientation(tmp_path):

    sub_images = []
    for idx, (pfile, dcm, pos) in enumerate(zip(mrsi_path, mrsi_dcm_path, mrsi_t1_pos)):
        subprocess.check_call(['spec2nii', 'ge',
                               '-f', 'mrsi',
                               '-o', tmp_path,
                               '-j',
                               str(pfile)])

        # Make fsleyes rendering
        subprocess.check_call(['fsleyes', 'render',
                               '-of', tmp_path / f'mrsi_{idx}.png',
                               '-vl', str(pos[0]), str(pos[1]), str(pos[2]),
                               '-xc', '0', '0', '-yc', '0', '0', '-zc', '0', '0',
                               '-hc', ge_path / 'from_dicom' / 'T1.nii.gz',
                               '-dr', '-211', '7400',
                               str(dcm), '-a', '50', '-cm', 'red',
                               tmp_path / 'mrsi.nii.gz', '-ot', 'complex',
                               '-a', '50', '-cm', 'blue'])

        fsl_ss = Image.open(tmp_path / f'mrsi_{idx}.png')
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = get_concat_h(Image.open(ss_path / f'mrsi_{idx + 1}.0001.jpg'),
                          Image.open(ss_path / f'mrsi_{idx + 1}.0002.jpg'),
                          Image.open(ss_path / f'mrsi_{idx + 1}.0003.jpg'))
        ss_cropped = ss  # ss.crop((0, 110, width - 50, height / 2 - 30))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 2))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * r, si.height * c))
        draw.text((10 + si.width * r, 10 + si.height * c), mrsi_path[idx].stem, (255, 0, 0))

    final_img.save(output_path / 'ge_mrsi.png')



'''OLD Orientation tests for Philips DICOM format conversion.
This test has been superseded by the new test test_philips_orientation_dcm_spar_new.py

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path

import pytest
from PIL import Image, ImageDraw

# Data paths
output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'

data_path = [philips_path / 'DICOM' / 'SV_phantom_center' / 'IM-0018-0002-0001.dcm',
             philips_path / 'DICOM' / 'SV_phantom_center_no_Water_Suppression' / 'IM-0023-0002-0001.dcm',
             philips_path / 'DICOM' / 'SV_phantom_H15mm' / 'IM-0020-0002-0001.dcm',
             philips_path / 'DICOM' / 'SV_phantom_R15mm' / 'IM-0019-0002-0001.dcm',
             philips_path / 'DICOM' / 'SV_phantom_45deg_RL' / 'IM-0022-0002-0001.dcm',
             philips_path / 'DICOM' / 'SV_phantom_45deg_AP' / 'IM-0021-0002-0001.dcm']

structural_data = philips_path / 'DICOM' / 'structural' / '1001_SmartBrain.nii.gz'

screenshot_path = [philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_center',
                   philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_center_no_Water_Suppression',
                   philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_H15mm',
                   philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_R15mm',
                   philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_45deg_RL',
                   philips_path / 'DICOM' / 'screenshots' / 'SV_phantom_45deg_AP']

t1_pos = [(134, 155, 56),
          (134, 155, 56),
          (134, 155, 56),
          (134, 155, 56),
          (134, 155, 56),
          (134, 155, 56)]


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def crop_and_flip_first_third_and_swap(img_in, swap='23'):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 180, int(width / 3) + 15, height - 180))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 16, 180, 2 * int(width / 3) + 16, height - 180))
    cropped_2 = img_in.crop((2 * int(width / 3) + 17, 180, width, height - 180))

    out = Image.new('RGB', (width, height - 360))
    if swap == '23':
        out.paste(cropped_0, (0, 0))
        out.paste(cropped_2, (int(width / 3) + 16, 0))
        out.paste(cropped_1, (2 * int(width / 3) + 17, 0))
    elif swap == '123':
        out.paste(cropped_2, (0, 0))
        out.paste(cropped_0, (int(width / 3), 0))
        out.paste(cropped_1, (2 * int(width / 3) + 17, 0))
    return out


@pytest.mark.orientation
def test_svs_orientation(tmp_path):
    sub_images = []
    for idx, (d_path, p_path, pos) in enumerate(zip(data_path, screenshot_path, t1_pos)):
        subprocess.check_call(['spec2nii', 'philips_dcm',
                               '-f', f'svs_{idx}',
                               '-o', tmp_path,
                               '-j',
                               str(d_path)])

        # Make fsleyes rendering
        subprocess.check_call(['fsleyes', 'render',
                               '-of', tmp_path / f'svs_{idx}.png',
                               '-vl', str(pos[0]), str(pos[1]), str(pos[2]),
                               '-xc', '0', '0', '-yc', '0', '0', '-zc', '0', '0',
                               '-hc', structural_data,
                               tmp_path / f'svs_{idx}.nii.gz', '-ot', 'complex',
                               '-a', '50', '-cm', 'blue'])

        fsl_ss = Image.open(tmp_path / f'svs_{idx}.png')
        width, height = fsl_ss.size
        if idx == 2:
            swap = '123'
        else:
            swap = '23'
        fsl_ss_cropped = crop_and_flip_first_third_and_swap(fsl_ss, swap=swap)

        ss = Image.open(p_path / 'placement.jpg')
        width, height = ss.size
        # (left, upper, right, lower)
        ss_cropped = ss.crop((10, 90, width - 300, height / 2 - 150))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 3))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * c, si.height * r))
        draw.text((10 + si.width * c, 10 + si.height * r), f'P{idx + 1}', (255, 0, 0))

    final_img.save(output_path / 'philips_dicom_svs.png')
    print(tmp_path)



"""spec2nii module containing functions specific to interpreting Siemens DICOM
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
from datetime import datetime
import re
import warnings

import numpy as np
import nibabel.nicom.dicomwrappers
from nibabel.nicom import csareader as csar

from mapvbvd.read_twix_hdr import parse_buffer

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.dcm2niiOrientation.orientationFuncs import dcm_to_nifti_orientation
from spec2nii import __version__ as spec2nii_ver


class inconsistentDataError(Exception):
    pass


class missingTagError(Exception):
    pass


class IncompatibleSOPClassUID(Exception):
    pass


def xa_or_vx(img):
    """Determine siemens DICOM type

    :param img: DICOM image object
    :type img: nibabel.nicom.dicomwrappers.SiemensWrapper
    :raises IncompatibleSOPClassUID: Raised if not a Siemens Syngo Non Image Storage or MRSpectroscopyStorage dicom file
    :return: String either 'xa' or 'vx'
    :rtype: str
    """
    if img.dcm_data.SOPClassUID == '1.3.12.2.1107.5.9.1':
        return 'vx'
    elif img.dcm_data.SOPClassUID == '1.2.840.10008.5.1.4.1.1.4.2':
        return 'xa'
    else:
        if img.dcm_data.SOPClassUID == '1.2.840.10008.5.1.4.1.1.4':
            raise IncompatibleSOPClassUID(
                f'spec2nii detected SOPClassUID {img.dcm_data.SOPClassUID}.'
                ' This normally contains MR imaging (not spectroscopy) data.'
                ' This data was collected on a'
                f' {img.dcm_data.SoftwareVersions} baseline scanner.'
                ' spec2nii is tested on VA-VE, XA20, and XA30 DICOM files.')
        else:
            raise IncompatibleSOPClassUID(
                'spec2nii does not recognise this SOPClassUID '
                f'{img.dcm_data.SOPClassUID} as MRS data. This data was collected on a'
                f' {img.dcm_data.SoftwareVersions} baseline scanner.'
                ' spec2nii is tested on VA-VE, XA20, and XA30 DICOM files.')


def svs_or_CSI(img):
    """Identify from the csa headers whether data is CSI or SVS."""
    if xa_or_vx(img) == 'vx':
        rows = img.csa_header['tags']['Rows']['items'][0]
        cols = img.csa_header['tags']['Columns']['items'][0]
        slices = img.csa_header['tags']['NumberOfFrames']['items'][0]
    elif xa_or_vx(img) == 'xa':
        spec_fov_geom = img.dcm_data.SharedFunctionalGroupsSequence[0].MRSpectroscopyFOVGeometrySequence[0]
        rows = spec_fov_geom.SpectroscopyAcquisitionPhaseRows
        cols = spec_fov_geom.SpectroscopyAcquisitionPhaseColumns
        slices = spec_fov_geom.SpectroscopyAcquisitionOutOfPlanePhaseSteps

    if np.prod([rows, cols, slices]) > 1.0:
        return 'CSI'
    else:
        return 'SVS'


def multi_file_dicom(files_in, fname_out, tag, verbose):
    """Parse a list of Siemens DICOM files"""

    # Convert each file (combine after)
    data_list = []
    orientation_list = []
    dwelltime_list = []
    meta_list = []
    series_num = []
    inst_num = []
    reference = []
    str_suffix = []
    mainStr = ''
    first_mrs_dcm_found = False
    for idx, fn in enumerate(files_in):
        if verbose:
            print(f'Converting dicom file {fn}')

        img = nibabel.nicom.dicomwrappers.wrapper_from_file(fn)

        try:
            mrs_type = svs_or_CSI(img)
        except IncompatibleSOPClassUID as exc:
            if len(files_in) == 1:
                raise exc
            else:
                print(f'Skipping {fn}.')
                print('Raised IncompatibleSOPClassUID error. Moving to next file.')
                print(f'Message: {str(exc)}\n')
                continue

        if mrs_type == 'SVS':
            specDataCmplx, orientation, dwelltime, meta_obj = process_siemens_svs(img, verbose=verbose)

            newshape = (1, 1, 1) + specDataCmplx.shape
            specDataCmplx = specDataCmplx.reshape(newshape)

        else:
            specDataCmplx, orientation, dwelltime, meta_obj = process_siemens_csi(img, verbose=verbose)

        data_list.append(specDataCmplx)
        orientation_list.append(orientation)
        dwelltime_list.append(dwelltime)
        meta_list.append(meta_obj)

        series_num.append(int(img.dcm_data.SeriesNumber))
        inst_num.append(int(img.dcm_data.InstanceNumber))

        ref_ind, str_suf = identify_integrated_references(img, img.dcm_data.InstanceNumber)
        reference.append(ref_ind)
        str_suffix.append(str_suf)

        if not first_mrs_dcm_found:
            first_mrs_dcm_found = True
            if fname_out:
                mainStr = fname_out
            elif 'SeriesDescription' in img.dcm_data:
                mainStr = img.dcm_data.SeriesDescription
            elif 'SeriesInstanceUID' in img.dcm_data:
                mainStr = img.dcm_data.SeriesInstanceUID
            else:
                raise missingTagError("Neither SeriesDescription or SeriesInstanceUID tags defined."
                                      " Please specify an output filename using '-f'")

    # Sort by series and instance number
    data_list = np.asarray(data_list)
    orientation_list = np.asarray(orientation_list)
    dwelltime_list = np.asarray(dwelltime_list)
    meta_list = np.asarray(meta_list)
    series_num = np.asarray(series_num)
    inst_num = np.asarray(inst_num)
    reference = np.asarray(reference)
    str_suffix = np.asarray(str_suffix)
    files_in = np.asarray(files_in)

    sort_index = np.lexsort((inst_num, series_num))  # Sort by series then inst

    data_list = data_list[sort_index, :]
    orientation_list = orientation_list[sort_index]
    dwelltime_list = dwelltime_list[sort_index]
    meta_list = meta_list[sort_index]
    series_num = series_num[sort_index]
    inst_num = inst_num[sort_index]
    reference = reference[sort_index]
    str_suffix = str_suffix[sort_index]
    files_in = files_in[sort_index]

    group_ind = []
    for sn in np.unique(series_num):
        for rn in np.unique(reference):
            group_ind.append(list(np.where(np.logical_and(series_num == sn, reference == rn))[0]))

    if verbose:
        print(f'Sorted series numbers: {series_num}')
        print(f'Sorted instance numbers: {inst_num}')
        print(f'Sorted reference index: {reference}')
        print(f'Output groups: {group_ind}')

    nifti_mrs_out, fnames_out = [], []
    for idx, gr in enumerate(group_ind):

        # If data shape, orientation, dwelltime match then
        # proceed
        def not_equal(lst):
            return lst[:-1] != lst[1:]

        if not_equal([d.shape for d in data_list[gr]])\
                and not_equal([o.Q44.tolist() for o in orientation_list[gr]])\
                and not_equal(dwelltime_list[gr]):
            raise inconsistentDataError('Shape, orientation and dwelltime must match in combined data.')

        fnames_out.append(mainStr + str_suffix[gr[0]])

        dt_used = dwelltime_list[gr[0]]
        or_used = orientation_list[gr[0]]

        # Add original files to nifti meta information.
        meta_used = meta_list[gr[0]]
        meta_used.set_standard_def('OriginalFile', [str(ff) for ff in files_in[gr]])

        # Combine data into 5th dimension if needed
        data_in_gr = data_list[gr]
        if len(data_in_gr) > 1:
            combined_data = np.stack(data_in_gr, axis=-1)
        else:
            combined_data = data_in_gr[0]

        # Add dimension information (if not None for default)
        if tag:
            meta_used.set_dim_info(0, tag)

        # Create NIFTI MRS object.
        try:
            nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(combined_data, dt_used, meta_used, or_used.Q44, no_conj=True))
        except np.linalg.LinAlgError:
            warnings.warn("""The quaternion passed to NIfTI_MRS was singular.
                           Most likely your slice position is not well defined. I have set it to default.""")
            nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(combined_data, dt_used, meta_used, no_conj=True))

    # If there are any identical names then append an index
    seen = np.unique(fnames_out)
    if seen.size < len(fnames_out):
        seen_count = np.zeros(seen.shape, dtype=int)
        fnames_out_checked = []
        for fn in fnames_out:
            if fn in seen:
                seen_index = seen == fn
                fnames_out_checked.append(fn + f'_{seen_count[seen_index][0]:03}')
                seen_count[seen_index] += 1
            else:
                fnames_out_checked.append(fn)

        return nifti_mrs_out, fnames_out_checked
    else:
        return nifti_mrs_out, fnames_out


def process_siemens_svs(img, verbose):
    """Pass to appropriate function for software version."""
    if xa_or_vx(img) == 'vx':
        return process_siemens_svs_vx(img, verbose)
    elif xa_or_vx(img) == 'xa':
        return process_siemens_svs_xa(img, verbose)


def _get_enhanced_dcm_img_orientation_pat(img):
    """Extract the image orientation patient field from an enhanced dicom scan

    :param img: DICOM image object
    :type img: nibabel.nicom.dicomwrappers.SiemensWrapper
    :return: 3x2 Numpy array containing the orientation info
    :rtype: np.ndarray
    """
    dcm_hdrs = img.dcm_data.PerFrameFunctionalGroupsSequence[0]
    dcm_hdrs1 = img.dcm_data.SharedFunctionalGroupsSequence[0]
    # 1) Extract dicom parameters
    try:
        return np.array(dcm_hdrs.PlaneOrientationSequence[0].ImageOrientationPatient).reshape(2, 3)
    except AttributeError:
        # For VB19 data formatted as MRSpectroscopyStorage this seems to be different to
        # the XA data I've seen. But this could also be a sequence specific thing!
        return np.array(dcm_hdrs1.PlaneOrientationSequence[0].ImageOrientationPatient).reshape(2, 3)


def process_siemens_svs_xa(img, verbose):
    """Process Siemens DICOM SVS data acquired on a NumarisX system."""
    specData = np.frombuffer(img.dcm_data[('5600', '0020')].value, dtype=np.single)
    # It appears the phase convention is reversed in the NumarisX systems.
    specDataCmplx = specData[0::2] - 1j * specData[1::2]

    dcm_hdrs = img.dcm_data.PerFrameFunctionalGroupsSequence[0]
    dcm_hdrs1 = img.dcm_data.SharedFunctionalGroupsSequence[0]
    # 1) Extract dicom parameters
    imageOrientationPatient = _get_enhanced_dcm_img_orientation_pat(img)
    imagePositionPatient = dcm_hdrs.PlanePositionSequence[0].ImagePositionPatient

    # The first two elements could easily be the reverse of this.
    xyzMM = np.array([float(dcm_hdrs1.PixelMeasuresSequence[0].PixelSpacing[0]),
                      float(dcm_hdrs1.PixelMeasuresSequence[0].PixelSpacing[1]),
                      float(dcm_hdrs1.PixelMeasuresSequence[0].SliceThickness)])

    currNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                    imagePositionPatient,
                                                    xyzMM,
                                                    (1, 1, 1),
                                                    verbose=verbose)
    dwelltime = 1 / img.dcm_data.SpectralWidth
    meta = extractDicomMetadata_xa(img)

    return specDataCmplx, currNiftiOrientation, dwelltime, meta


def process_siemens_svs_vx(img, verbose):
    """Process Siemens DICOM SVS data acquired on a Numaris4 Vx system."""

    specData = np.frombuffer(img.dcm_data[('7fe1', '1010')].value, dtype=np.single)
    specDataCmplx = specData[0::2] + 1j * specData[1::2]

    # 1) Extract dicom parameters
    imageOrientationPatient = np.array(img.csa_header['tags']['ImageOrientationPatient']['items']).reshape(2, 3)
    # VoiPosition - this does not have the FOV shift that imagePositionPatient has
    imagePositionPatient = img.csa_header['tags']['VoiPosition']['items']
    xyzMM = np.array([img.csa_header['tags']['VoiPhaseFoV']['items'][0],
                      img.csa_header['tags']['VoiReadoutFoV']['items'][0],
                      img.csa_header['tags']['VoiThickness']['items'][0]])

    currNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                    imagePositionPatient,
                                                    xyzMM,
                                                    (1, 1, 1),
                                                    verbose=verbose)
    dwelltime = img.csa_header['tags']['RealDwellTime']['items'][0] * 1E-9
    meta = extractDicomMetadata_vx(img)

    return specDataCmplx, currNiftiOrientation, dwelltime, meta


def process_siemens_csi(img, verbose):
    """Pass to appropriate CSI function for software version."""
    if xa_or_vx(img) == 'vx':
        return process_siemens_csi_vx(img, verbose)
    elif xa_or_vx(img) == 'xa':
        return process_siemens_csi_xa(img, verbose)


def process_siemens_csi_xa(img, verbose):
    """Process Siemens DICOM CSI data acquired on a NumarisX system or Numaris4 Vx using enhanced DCM export"""

    warnings.warn(
        'The orientation of CSI stored in enhanced DICOM remains weakly tested.'
        'Please provide more test data.')

    specData = np.frombuffer(img.dcm_data[('5600', '0020')].value, dtype=np.single)
    # It appears the phase convention is reversed in the NumarisX systems.
    specDataCmplx = specData[0::2] - 1j * specData[1::2]

    # The data in img.dcm_data.SharedFunctionalGroupsSequence[0].MRSpectroscopyFOVGeometrySequence[0]
    # is the uninterpolated (lower) resolution that doesn't match the data size
    rows = img.dcm_data.Rows
    cols = img.dcm_data.Columns
    slices = int(img.dcm_data.NumberOfFrames)
    spectral_points = img.dcm_data.DataPointColumns

    specDataCmplx = specDataCmplx.reshape((slices, rows, cols, spectral_points))
    specDataCmplx = np.moveaxis(specDataCmplx, (0, 1, 2), (2, 1, 0))

    dcm_hdrs = img.dcm_data.PerFrameFunctionalGroupsSequence[0]
    dcm_hdrs1 = img.dcm_data.SharedFunctionalGroupsSequence[0]

    # 1) Extract dicom parameters
    imageOrientationPatient = _get_enhanced_dcm_img_orientation_pat(img)
    # VoiPosition - this does not have the FOV shift that imagePositionPatient has
    imagePositionPatient = dcm_hdrs.PlanePositionSequence[0].ImagePositionPatient

    # The first two elements could easily be the reverse of this.
    xyzMM = np.array([float(dcm_hdrs1.PixelMeasuresSequence[0].PixelSpacing[0]),
                      float(dcm_hdrs1.PixelMeasuresSequence[0].PixelSpacing[1]),
                      float(dcm_hdrs1.PixelMeasuresSequence[0].SliceThickness)])

    # Note that half_shift = False. This is the opposite as for classic/original DCM.
    # THIS NEEDS TESTING AND DOCUMENTING. For an explanation of original shift see spec2nii/notes/siemens.md
    currNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                    imagePositionPatient,
                                                    xyzMM,
                                                    specDataCmplx.shape[:3],
                                                    half_shift=False,
                                                    verbose=verbose)

    dwelltime = 1 / img.dcm_data.SpectralWidth
    meta = extractDicomMetadata_xa(img)

    # Look for a VOI in the data
    meta = _detect_and_fill_voi_enh(img, meta)

    return specDataCmplx, currNiftiOrientation, dwelltime, meta


def process_siemens_csi_vx(img, verbose):
    """Process Siemens DICOM CSI data acquired on a Numaris4 Vx system."""
    specData = np.frombuffer(img.dcm_data[('7fe1', '1010')].value, dtype=np.single)
    specDataCmplx = specData[0::2] + 1j * specData[1::2]

    rows = img.csa_header['tags']['Rows']['items'][0]
    cols = img.csa_header['tags']['Columns']['items'][0]
    slices = img.csa_header['tags']['NumberOfFrames']['items'][0]
    spectral_points = img.csa_header['tags']['DataPointColumns']['items'][0]

    # I think the above is sufficient, but might need to access more specific Siemens specific parameters
    # In that case I could use the following use the full MrPhoenixProtocol sSpecPara parameters
    # fullcsa = csar.get_csa_header(img.dcm_data, csa_type='series')
    # xprot = parse_buffer(fullcsa['tags']['MrPhoenixProtocol']['items'][0])

    # rows = int(xprot[('sSpecPara', 'lFinalMatrixSizePhase')])
    # cols = int(xprot[('sSpecPara', 'lFinalMatrixSizeRead')])
    # slices = int(xprot[('sSpecPara', 'lFinalMatrixSizeSlice')])
    # spectral_points = int(xprot[('sSpecPara', 'lVectorSize')])

    specDataCmplx = specDataCmplx.reshape((slices, rows, cols, spectral_points))
    specDataCmplx = np.moveaxis(specDataCmplx, (0, 1, 2), (2, 1, 0))

    # 1) Extract dicom parameters
    imageOrientationPatient = np.array(img.csa_header['tags']['ImageOrientationPatient']['items']).reshape(2, 3)
    imagePositionPatient = np.array(img.csa_header['tags']['ImagePositionPatient']['items'])
    xyzMM = np.array([img.csa_header['tags']['PixelSpacing']['items'][0],
                      img.csa_header['tags']['PixelSpacing']['items'][1],
                      img.csa_header['tags']['SliceThickness']['items'][0]])

    # Note that half_shift = True. For an explanation see spec2nii/notes/siemens.md
    currNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                    imagePositionPatient,
                                                    xyzMM,
                                                    specDataCmplx.shape[:3],
                                                    half_shift=True,
                                                    verbose=verbose)

    dwelltime = img.csa_header['tags']['RealDwellTime']['items'][0] * 1E-9
    meta = extractDicomMetadata_vx(img)

    # Look for a VOI in the data
    meta = _detect_and_fill_voi(img, meta)

    return specDataCmplx, currNiftiOrientation, dwelltime, meta


def _detect_and_fill_voi(img, meta):
    """Look for a VOI volume in the headers of a CSI scan. If present populate the VOI header field.

    :param img: Nibable DICOM image object
    :type img: SiemensWrapper
    :param meta: Existing NIfTI-MRS meta object
    :type meta: hdr_ext
    :return: Modified meta hdr_ext object
    :rtype: hdr_ext
    """
    imageOrientationPatient = np.array(img.csa_header['tags']['ImageOrientationPatient']['items']).reshape(2, 3)
    # VoiPosition - this does not have the FOV shift that imagePositionPatient has
    imagePositionPatient = img.csa_header['tags']['VoiPosition']['items']
    xyzMM = np.array([img.csa_header['tags']['VoiPhaseFoV']['items'][0],
                      img.csa_header['tags']['VoiReadoutFoV']['items'][0],
                      img.csa_header['tags']['VoiThickness']['items'][0]])

    voiNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                   imagePositionPatient,
                                                   xyzMM,
                                                   (1, 1, 1))

    meta.set_standard_def('VOI', voiNiftiOrientation.Q44.tolist())

    return meta


def _detect_and_fill_voi_enh(img, meta):
    """Look for a VOI volume in the headers of a CSI scan stored in enhanced dicom.
    If present populate the VOI header field.

    :param img: Nibable DICOM image object
    :type img: SiemensWrapper
    :param meta: Existing NIfTI-MRS meta object
    :type meta: hdr_ext
    :return: Modified meta hdr_ext object
    :rtype: hdr_ext
    """

    imageOrientationPatient = _get_enhanced_dcm_img_orientation_pat(img)

    # VoiPosition - this does not have the FOV shift that imagePositionPatient has
    vol_loc_seq = img.dcm_data.VolumeLocalizationSequence
    imagePositionPatient = vol_loc_seq[0].MidSlabPosition

    # This order needs validating!!
    xyzMM = np.array([vol_loc_seq[1].SlabThickness,
                      vol_loc_seq[2].SlabThickness,
                      vol_loc_seq[0].SlabThickness])

    voiNiftiOrientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                                   imagePositionPatient,
                                                   xyzMM,
                                                   (1, 1, 1))

    meta.set_standard_def('VOI', voiNiftiOrientation.Q44.tolist())

    return meta


def extractDicomMetadata_xa(dcmdata):
    """ Extract information from the nibabel DICOM object to insert into the json header ext.
    For Numarix X systems

    Args:
        dcmdata: nibabel.nicom image object
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    dcm_hdrs1 = dcmdata.dcm_data.SharedFunctionalGroupsSequence[0]

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        dcmdata.dcm_data.TransmitterFrequency,
        dcmdata.dcm_data.ResonantNucleus)

    # Standard defined metadata
    def set_standard_def(nifti_mrs_key, location, key, cast=None):
        try:
            if cast is not None:
                obj.set_standard_def(nifti_mrs_key, cast(getattr(location, key)))
            else:
                obj.set_standard_def(nifti_mrs_key, getattr(location, key))
        except AttributeError:
            pass

    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', float(dcm_hdrs1.MREchoSequence[0].EffectiveEchoTime * 1E-3))
    # 'RepetitionTime'
    obj.set_standard_def(
        'RepetitionTime',
        float(dcm_hdrs1.MRTimingAndRelatedParametersSequence[0].RepetitionTime) / 1E3)
    # 'InversionTime'
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    obj.set_standard_def(
        'ExcitationFlipAngle',
        float(dcm_hdrs1.MRTimingAndRelatedParametersSequence[0].FlipAngle))
    # 'TxOffset'
    # 'VOI'
    # 'WaterSuppressed'
    # 'WaterSuppressionType'
    # 'SequenceTriggered'

    # # 5.2 Scanner information
    # 'Manufacturer'
    set_standard_def('Manufacturer', dcmdata.dcm_data, 'Manufacturer')
    # 'ManufacturersModelName'
    set_standard_def('ManufacturersModelName', dcmdata.dcm_data, 'ManufacturerModelName')
    # 'DeviceSerialNumber'
    set_standard_def('DeviceSerialNumber', dcmdata.dcm_data, 'DeviceSerialNumber', cast=str)
    # 'SoftwareVersions'
    set_standard_def('SoftwareVersions', dcmdata.dcm_data, 'SoftwareVersions')
    # 'InstitutionName'
    set_standard_def('InstitutionName', dcmdata.dcm_data, 'InstitutionName')
    # 'InstitutionAddress'
    set_standard_def('InstitutionAddress', dcmdata.dcm_data, 'InstitutionAddress')
    # 'TxCoil'
    obj.set_standard_def(
        'TxCoil',
        dcm_hdrs1.MRTransmitCoilSequence[0].TransmitCoilName)
    # 'RxCoil'
    obj.set_standard_def(
        'RxCoil',
        dcm_hdrs1.MRReceiveCoilSequence[0].ReceiveCoilName)

    # # 5.3 Sequence information
    # 'SequenceName'
    set_standard_def('SequenceName', dcmdata.dcm_data, 'PulseSequenceName')
    # 'ProtocolName'
    set_standard_def('ProtocolName', dcmdata.dcm_data, 'ProtocolName')
    # # 5.4 Sequence information
    # 'PatientPosition'
    set_standard_def('PatientPosition', dcmdata.dcm_data, 'PatientPosition')
    # 'PatientName'
    set_standard_def('PatientName', dcmdata.dcm_data.PatientName, 'family_name')
    # 'PatientID'
    set_standard_def('PatientID', dcmdata.dcm_data, 'PatientID')
    # 'PatientWeight'
    set_standard_def('PatientWeight', dcmdata.dcm_data, 'PatientWeight', cast=float)
    # 'PatientDoB'
    set_standard_def('PatientDoB', dcmdata.dcm_data, 'PatientBirthDate')
    # 'PatientSex'
    set_standard_def('PatientSex', dcmdata.dcm_data, 'PatientSex')

    # # 5.5 Provenance and conversion metadata
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)

    # 'OriginalFile'
    # Set elsewhere

    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    return obj


def extractDicomMetadata_vx(dcmdata):
    """ Extract information from the nibabel DICOM object to insert into the json header ext.

    Args:
        dcmdata: nibabel.nicom image object
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        dcmdata.csa_header['tags']['ImagingFrequency']['items'][0],
        dcmdata.csa_header['tags']['ImagedNucleus']['items'][0])

    # Standard defined metadata
    def set_standard_def(nifti_mrs_key, location, key, cast=None):
        try:
            if cast is not None:
                obj.set_standard_def(nifti_mrs_key, cast(getattr(location, key)))
            else:
                obj.set_standard_def(nifti_mrs_key, getattr(location, key))
        except AttributeError:
            pass

    # # 5.1 MRS specific Tags
    # 'EchoTime' - requires substantial extraction from the full protocol in case there are multiple
    # sub-values, e.g. summing the three TEs of a sLAASER sequence together.
    fullcsa = csar.get_csa_header(dcmdata.dcm_data, csa_type='series')
    xprot = parse_buffer(fullcsa['tags']['MrPhoenixProtocol']['items'][0])
    te_sum = 0
    for teDx in range(128):
        try:
            te_sum += xprot[('alTE', str(teDx))]
        except KeyError:
            break
    obj.set_standard_def('EchoTime', float(te_sum * 1E-6))
    # Original
    # obj.set_standard_def('EchoTime', float(dcmdata.csa_header['tags']['EchoTime']['items'][0] * 1E-3))

    # 'RepetitionTime'
    obj.set_standard_def('RepetitionTime', float(dcmdata.csa_header['tags']['RepetitionTime']['items'][0] / 1E3))
    # 'InversionTime'
    if dcmdata.csa_header['tags']['InversionTime']['n_items'] > 0:
        obj.set_standard_def('InversionTime', float(dcmdata.csa_header['tags']['InversionTime']['items'][0]))
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    obj.set_standard_def('ExcitationFlipAngle', float(dcmdata.csa_header['tags']['FlipAngle']['items'][0]))
    # 'TxOffset'
    # 'VOI'
    # 'WaterSuppressed'
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    set_standard_def('Manufacturer', dcmdata.dcm_data, 'Manufacturer')
    # 'ManufacturersModelName'
    set_standard_def('ManufacturersModelName', dcmdata.dcm_data, 'ManufacturerModelName')
    # 'DeviceSerialNumber'
    set_standard_def('DeviceSerialNumber', dcmdata.dcm_data, 'DeviceSerialNumber', cast=str)
    # 'SoftwareVersions'
    set_standard_def('SoftwareVersions', dcmdata.dcm_data, 'SoftwareVersions')
    # 'InstitutionName'
    set_standard_def('InstitutionName', dcmdata.dcm_data, 'InstitutionName')
    # 'InstitutionAddress'
    set_standard_def('InstitutionAddress', dcmdata.dcm_data, 'InstitutionAddress')
    # 'TxCoil'
    obj.set_standard_def(
        'TxCoil',
        dcmdata.csa_header['tags']['TransmittingCoil']['items'][0])
    # 'RxCoil'
    if len(dcmdata.csa_header['tags']['ReceivingCoil']['items']) > 0:
        obj.set_standard_def('RxCoil',
                             dcmdata.csa_header['tags']['ReceivingCoil']['items'][0])
    else:
        obj.set_standard_def('RxCoil',
                             dcmdata.csa_header['tags']['ImaCoilString']['items'][0])
    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', dcmdata.csa_header['tags']['SequenceName']['items'][0])
    # 'ProtocolName'
    set_standard_def('ProtocolName', dcmdata.dcm_data, 'ProtocolName')
    # # 5.4 Sequence information
    # 'PatientPosition'
    set_standard_def('PatientPosition', dcmdata.dcm_data, 'PatientPosition')
    # 'PatientName'
    set_standard_def('PatientName', dcmdata.dcm_data.PatientName, 'family_name')
    # 'PatientID'
    # 'PatientWeight'
    set_standard_def('PatientWeight', dcmdata.dcm_data, 'PatientWeight', cast=float)
    # 'PatientDoB'
    set_standard_def('PatientDoB', dcmdata.dcm_data, 'PatientBirthDate')
    # 'PatientSex'
    set_standard_def('PatientSex', dcmdata.dcm_data, 'PatientSex')
    # # 5.5 Provenance and conversion metadata
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    # Set elsewhere
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    # Some additional sequence information
    obj.set_user_def(key='PulseSequenceFile',
                     value=dcmdata.csa_header['tags']['SequenceName']['items'][0],
                     doc='Sequence binary path.')
    # obj.set_user_def(key='IceProgramFile',
    #                  value=mapVBVDHdr['Meas'][('tICEProgramName')],
    #                  doc='Reconstruction binary path.')

    return obj


def identify_integrated_references(img, inst_num):
    '''Heuristics for identifying integrated reference scans in known sequences.
    Sequences handled: CMRR svs_slaserVOI_dkd(2)

    :param img: nibable dicom image

    :return: Ref scan index 0 = not reference scan, higher integer splits into groups.
    :return: name suffix
    '''

    if xa_or_vx(img) == 'xa':
        return 0, ''

    fullcsa = csar.get_csa_header(img.dcm_data, csa_type='series')
    xprot = parse_buffer(fullcsa['tags']['MrPhoenixProtocol']['items'][0])

    # Handle CMRR DKD sequence
    # https://www.cmrr.umn.edu/spectro/
    # SEMI-LASER (MRM 2011, NMB 2019) Release 2016-12
    # Three cases as explained by DKD.
    '''
    dkd_slaserVOI_dkd
        ScanMode = 8
        (NOTE DKD originally said 2, but from testing the sLASER_VE11C_Sept2016 version 8 is the only option
        xprot[('sSpecPara', 'lAutoRefScanMode')] = 1 when set to 'off' and 8 when set to 'on')
            4 water ref at start and 4 at the end where the first 2 water ref in
            each are acquired with VAPOR RF off but OVS on while last 2 ref are with VAPOR and OVS completely off.

    dkd_slaserVOI_dkd2
        ScanMode =8
            does the same as ScanMode =8 above

        ScanMode =2 with ScanNo=2
            this means 2 water ref at start and 2 at end;
            this acq is done with VAPOR and OVS RF off, but all gradients still ON.
    '''
    seq_file_name = xprot[('tSequenceFileName',)].strip('"').lower()
    if (re.search(r'svs_slaser(voi)?_dkd2$', seq_file_name)
        and xprot[('sSpecPara', 'lAutoRefScanMode')] == 8.0)\
        or (re.search(r'svs_slaser(voi)?_dkd$', seq_file_name)
            and xprot[('sSpecPara', 'lAutoRefScanMode')] == 8.0):
        num_ref = int(xprot[('sSpecPara', 'lAutoRefScanNo')])
        num_dyn = int(xprot[('lAverages',)])
        total_dyn = num_dyn + (num_ref * 4)
        if inst_num <= num_ref:
            # First ecc calibration references
            return 1, '_rf_off'
        elif inst_num <= (num_ref * 2):
            # First quantitation calibration references
            return 2, '_rf_grads_ovs_off'
        elif (total_dyn - (2 * num_ref)) < inst_num <= (total_dyn - num_ref):
            # Second ecc calibration references
            return 1, '_rf_off'
        elif (total_dyn - num_ref) < inst_num <= total_dyn:
            # Second quantitation calibration references
            return 2, '_rf_grads_ovs_off'
        else:
            return 0, ''
    elif (re.search(r'svs_slaser(voi)?_dkd2', seq_file_name)
            and xprot[('sSpecPara', 'lAutoRefScanMode')] == 2.0):
        num_ref = int(xprot[('sSpecPara', 'lAutoRefScanNo')])
        num_dyn = int(xprot[('lAverages',)])
        total_dyn = num_dyn + (num_ref * 2)
        if inst_num < num_ref:
            # First WS and OVS RF off
            return 1, '_vapor_ovs_rfoff'
        elif inst_num >= (total_dyn - num_ref):
            # SecondWS and OVS RF off
            return 1, '_vapor_ovs_rfoff'
        else:
            return 0, ''
    else:
        return 0, ''



import numpy as np
import os.path as op
import subprocess
from PIL import Image, ImageDraw
from .io_for_tests import read_nifti_mrs
import pytest
from pathlib import Path

# Data paths
output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

siemens_path = op.join(op.dirname(__file__), 'spec2nii_test_data', 'Siemens')
vb_path = op.join(siemens_path, 'VBData')
ve_path = op.join(siemens_path, 'VEData')

# SVS VB data
svs_data_names_vb = ['C_T15_S10_10',
                     'S_C10_S5_10',
                     'T_C15_S10_10',
                     'iso_tra']

svs_data_twix_vb = ['Twix/meas_MID151_svs_se_C_T15_S10_10_FID108741.dat',
                    'Twix/meas_MID149_svs_se_S_C10_S5_10_FID108739.dat',
                    'Twix/meas_MID147_svs_se_T_C15_0_S10_0_10_FID108737.dat',
                    'Twix/meas_MID153_svs_se_Tra_sat_FID108743.dat']

svs_Data_dicom_vb = ['DICOM/svs_se_C>T15>S10_10_12_1',
                     'DICOM/svs_se_S>C10>S5_10_11_1',
                     'DICOM/svs_se_T>C15.0>S10.0_10_10_1',
                     'DICOM/svs_se_Tra_sat_13_1']

screen_shots_vb = ['Screenshots/svs_se_C_T15_S10_10.png',
                   'Screenshots/svs_se_S_C10_S5_10.png',
                   'Screenshots/svs_se_T_C15_0_S10_0_10.png',
                   'Screenshots/svs_se_Tra_sat.png']

# SVS VE data
svs_data_names_ve = ['C_T15_S10_10',
                     'S_C10_S5_10',
                     'T_C15_S10_10',
                     'iso_tra']

svs_data_twix_ve = ['Twix/meas_MID00240_FID62745_svs_se_c_t15_s10_R10.dat',
                    'Twix/meas_MID00238_FID62743_svs_se_s_c10_t5_R10.dat',
                    'Twix/meas_MID00235_FID62740_svs_se_t_c15_s10_R10.dat',
                    'Twix/meas_MID00242_FID62747_svs_se_iso_tra_sat.dat']

svs_Data_dicom_ve = ['DICOM/svs_se_c>t15>s10_R10_12_1',
                     'DICOM/svs_se_s>c10>t5_R10_11_1',
                     'DICOM/svs_se_t>c15>s10_R10_10_1',
                     'DICOM/svs_se_iso_tra_sat_13_1']

screen_shots_ve = ['Screenshots/svs_se_c_t15_s10_R10.png',
                   'Screenshots/svs_se_s_c10_t5_R10.png',
                   'Screenshots/svs_se_t_c15_s10_R10.png',
                   'Screenshots/svs_se_iso_tra_sat.png']


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def crop_and_flip_first_third(img_in):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 180, int(width / 3), height - 180))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 1, 180, width, height - 180))
    out = Image.new('RGB', (width, height - 360))
    out.paste(cropped_0, (0, 0))
    out.paste(cropped_1, (int(width / 3) + 1, 0))
    return out


# Test siemens VB svs
@pytest.mark.orientation
def test_VB(tmp_path):
    sub_images = []
    for idx, (f_t, f_d, name) in enumerate(zip(svs_data_twix_vb, svs_Data_dicom_vb, svs_data_names_vb)):
        # Convert twix
        subprocess.check_call(['spec2nii', 'twix',
                               '-e', 'image',
                               '-f', name + '_t',
                               '-o', tmp_path,
                               '-j', op.join(vb_path, f_t)])
        # Convert DICOM
        subprocess.check_call(['spec2nii', 'dicom',
                               '-f', name + '_d',
                               '-o', tmp_path,
                               '-j', op.join(vb_path, f_d)])

        # Make fsleyes rendering
        subprocess.check_call(['fsleyes',
                               'render', '-of', op.join(tmp_path, f'svs_{idx}.png'),
                               '-vl', '95', '89', '96',
                               '-hc', op.join(vb_path, 'T1.nii.gz'),
                               op.join(tmp_path, name + '_t.nii.gz'),
                               '-ot', 'complex', '-a', '50', '-cm', 'red',
                               op.join(tmp_path, name + '_d.nii.gz'),
                               '-ot', 'complex', '-a', '50', '-cm', 'blue'])

        img_t = read_nifti_mrs(op.join(tmp_path, name + '_t.nii.gz'))
        img_d = read_nifti_mrs(op.join(tmp_path, name + '_d.nii.gz'))

        assert np.allclose(img_t.affine, img_d.affine)

        fsl_ss = Image.open(op.join(tmp_path, f'svs_{idx}.png'))
        width, height = fsl_ss.size
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = Image.open(op.join(vb_path, screen_shots_vb[idx]))
        width, height = ss.size
        ss_cropped = ss.crop((0, 50, width - 50, height / 2 - 80))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 2))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * r, si.height * c))
        draw.text((10 + si.width * r, 10 + si.height * c), svs_data_names_vb[idx], (255, 0, 0))

    final_img.save(output_path / 'svs_vb.png')


@pytest.mark.orientation
def test_VE(tmp_path):
    sub_images = []
    for idx, (f_t, f_d, name) in enumerate(zip(svs_data_twix_ve, svs_Data_dicom_ve, svs_data_names_ve)):
        # Convert twix
        subprocess.check_call(['spec2nii', 'twix',
                               '-e', 'image',
                               '-f', name + '_t',
                               '-o', tmp_path,
                               '-j', op.join(ve_path, f_t)])
        # Convert DICOM
        subprocess.check_call(['spec2nii', 'dicom',
                               '-f', name + '_d',
                               '-o', tmp_path,
                               '-j', op.join(ve_path, f_d)])

        # Make fsleyes rendering
        if idx == 3:
            x, y, z = '57', '48', '37'
        else:
            x, y, z = '53', '67', '56'
        subprocess.check_call(['fsleyes',
                               'render', '-of', op.join(tmp_path, f'svs_{idx}.png'),
                               '-vl', x, y, z,
                               '-hc', op.join(ve_path, 'T1.nii.gz'),
                               op.join(tmp_path, name + '_t.nii.gz'),
                               '-ot', 'complex', '-a', '50', '-cm', 'red',
                               op.join(tmp_path, name + '_d.nii.gz'),
                               '-ot', 'complex', '-a', '50', '-cm', 'blue'])

        img_t = read_nifti_mrs(op.join(tmp_path, name + '_t.nii.gz'))
        img_d = read_nifti_mrs(op.join(tmp_path, name + '_d.nii.gz'))

        assert np.allclose(img_t.affine, img_d.affine)

        fsl_ss = Image.open(op.join(tmp_path, f'svs_{idx}.png'))
        width, height = fsl_ss.size
        fsl_ss_cropped = crop_and_flip_first_third(fsl_ss)

        ss = Image.open(op.join(ve_path, screen_shots_ve[idx]))
        width, height = ss.size
        ss_cropped = ss.crop((0, 50, width - 50, height / 2 - 80))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 2))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * r, si.height * c))
        draw.text((10 + si.width * r,
                  10 + si.height * c),
                  svs_data_names_ve[idx],
                  (255, 0, 0))

    final_img.save(output_path / 'svs_ve.png')



'''Tests for Varian format conversion.

Copyright Jack Miller, University of Oxford and University of Aarhus, 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
varian_path = Path(__file__).parent / 'spec2nii_test_data' / 'varian'
data_path = varian_path / '13C_spuls_01.fid'


def test_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'varian',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           '-d',
                           '-t', 'DIM_USER_0',
                           str(data_path)])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048, 1, 80)
    assert np.iscomplexobj(img_t.dataobj)
    assert np.isclose(1 / img_t.header['pixdim'][4], 40323, atol=2)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 100.4451626
    assert hdr_ext['ResonantNucleus'][0] == "13C"
    assert hdr_ext['OriginalFile'][0] == data_path.name
    assert 'VarianProcpar' in hdr_ext
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_USER_0'



"""spec2nii module containing functions specific to interpreting jmrui formats
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
import re
from datetime import datetime

import numpy as np

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.nifti_orientation import NIFTIOrient
from spec2nii import __version__ as spec2nii_ver


default_nuc_index = [None, "1H", "31P", "13C", "19F", "23NA"]
nuclei = ['1H', '2H', '3H', '13C', '23NA', '31P', '129XE']
gyro_ratios = [42.577, 6.536, 45.415, 10.7084, 11.262, 17.235, -11.777]


def id_nucleus(nucleus_hdr, trans_freq, b0):
    """Identify the nucleus string. TypeOfNucleus is a header key but
    some files seem to not have it in and sometimes it is an int

    :param: nucleus_hdr value of the TypeOfNucleus header key
    :param trans_freq: Transmitter/central frequency
    :param b0: Field strength in tesla.
    """

    if nucleus_hdr is not None\
            and isinstance(nucleus_hdr, str):
        return nucleus_hdr
    elif nucleus_hdr is not None\
            and isinstance(nucleus_hdr, (int, float))\
            and int(nucleus_hdr) > -1\
            and default_nuc_index[int(nucleus_hdr)] is not None:
        return default_nuc_index[int(nucleus_hdr)]
    else:
        try:
            gyrom_mag_r = trans_freq / b0
        except ZeroDivisionError:
            print('Cannot identify nucleus, using 1H default.')
            return '1H'

        if gyrom_mag_r > 1E4:
            gyrom_mag_r /= 1E6
        for nn, gr in zip(nuclei, gyro_ratios):
            if np.isclose(gr, gyrom_mag_r, atol=5E0):
                return nn

        print('Cannot identify nucleus, using 1H default.')
        return '1H'


def jmrui_format(args):
    '''Identify if text or binary format jmrui.'''
    if args.file.suffix == '.txt':
        return jmrui_txt(args)
    elif args.file.suffix == '.mrui':
        return jmrui_mrui(args)
    else:
        raise ValueError('jMRUI file format must be .txt or .mrui format.')


def jmrui_mrui(args):
    '''Process .mrui format files.'''

    data, header, str_info = read_mrui(args.file)

    newshape = (1, 1, 1) + data.shape
    data = data.reshape(newshape)

    # meta-data
    dwelltime = header['sampling_interval'] * 1E-3

    meta = jmrui_hdr_to_obj_mrui(header, str_info)
    meta.set_standard_def('OriginalFile', [args.file.name, ])

    if data.ndim > 4:
        meta.set_dim_info(0, 'DIM_USER_0', info='jMRUI frames')

    # Read optional affine file
    if args.affine:
        affine = np.loadtxt(args.affine)
    else:
        tmp = np.array([10000, 10000, 10000, 1])
        affine = np.diag(tmp)

    nifti_orientation = NIFTIOrient(affine)

    img_out = [gen_nifti_mrs_hdr_ext(
        data,
        dwelltime,
        meta,
        nifti_orientation.Q44,
        no_conj=True), ]

    # File names
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [args.file.stem, ]

    # Place in data output format
    return img_out, fname_out


def read_mrui(file_path):
    '''Read the header and data from a .mrui format file.

    :param Path file: file_path pathlib Path object to .mrui file.

    :return: Data in numpy format (npoints x nframes)
    :return: header dict
    :return: Additional information sorted as strings in file
    '''
    with open(file_path, 'br') as fp:
        # Header and data is stored as big endian double
        dt = np.dtype(np.float64)
        dt = dt.newbyteorder('>')

        # Read header values
        hdr = np.fromfile(fp, dtype=dt, count=13, offset=0)

        # Read data
        fp.seek(0, 2)
        file_size = fp.tell()
        n_frames = int(np.floor((file_size - 512) / (8 * hdr[1] * 2)))

        fp.seek(512, 0)
        data_count = int(n_frames * hdr[1] * 2)
        data = np.fromfile(fp, dtype=dt, count=data_count, offset=0)

        # Read final string info
        file_str = fp.read().decode("utf-8")

    # Sort header information
    header = {
        'type_of_sig': hdr[0],
        'number_of_points': int(hdr[1]),
        'sampling_interval': hdr[2],
        'begin_time ': hdr[3],
        'zero_order_phs': hdr[4],
        'transmitter_frequency': hdr[5],
        'magnetic_field': hdr[6],
        'type_of_nucleus': hdr[7],
        'reference_frequency_hz': hdr[8],
        'reference_frequency_ppm': hdr[9],
        'fid_or_echo': hdr[10],
        'apodizing': hdr[11],
        'num_zeros_view': hdr[12],
    }

    data = data[0::2] + 1j * data[1::2]

    data = data.reshape((n_frames, header['number_of_points'])).T
    data = data.squeeze()
    data = data.conj()

    return data, header, file_str


def jmrui_hdr_to_obj_mrui(header, str_info):
    """Translate jMRUI mrui header to NIfTI MRS"""

    nucleus = id_nucleus(header['type_of_nucleus'],
                         header['transmitter_frequency'],
                         header['magnetic_field'])

    meta = Hdr_Ext(
        header['transmitter_frequency'],
        nucleus)

    # meta.set_standard_def('ManufacturersModelName', header['Spectrometer'])
    # meta.set_standard_def('PatientName', header['NameOfPatient'])
    meta.set_standard_def('TxOffset', header['reference_frequency_ppm'])

    meta.set_standard_def('ConversionMethod', 'spec2nii')
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)

    meta.set_user_def(key='AdditionalInformation',
                      value=str_info,
                      doc='jMRUI .mrui string field.')

    return meta


def jmrui_txt(args):
    """Function for handling jmrui text files."""

    data, header = readjMRUItxt(args.file)

    newshape = (1, 1, 1) + data.shape
    data = data.reshape(newshape)

    # meta-data
    dwelltime = header['SamplingInterval'] * 1E-3

    meta = jmrui_hdr_to_obj(header)
    meta.set_standard_def('OriginalFile', [args.file.name, ])

    if data.ndim > 4:
        meta.set_dim_info(0, 'DIM_USER_0', info='jMRUI signals')

    # Read optional affine file
    if args.affine:
        affine = np.loadtxt(args.affine)
    else:
        tmp = np.array([10000, 10000, 10000, 1])
        affine = np.diag(tmp)

    nifti_orientation = NIFTIOrient(affine)

    img_out = [gen_nifti_mrs_hdr_ext(
        data,
        dwelltime,
        meta,
        nifti_orientation.Q44,
        no_conj=True), ]

    # File names
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [args.file.stem, ]

    # Place in data output format
    return img_out, fname_out


def readjMRUItxt(filename):
    """
    Read .txt format file
    Parameters
    ----------
    filename : string
        Name of jmrui .txt file

    Returns
    -------
    array-like
        Complex data
    dict
        Header information

    """
    signalRe = re.compile(r'Signal (?:number: |)(\d{1,}) out of (\d{1,}) in file')
    headerRe = re.compile(r'(\w*):(.*)')
    header = {}
    data   = []
    recordData = False
    signal_index = 0
    with open(filename, 'r') as txtfile:
        for line in txtfile:
            headerComp = headerRe.match(line)
            if headerComp:
                value = headerComp[2].strip()
                header.update({headerComp[1]: num(value)})

            signalIndices = signalRe.match(line)
            if signalIndices:
                recordData = True
                signal_index += 1
                continue

            if recordData:
                curr_data = line.split()
                if len(curr_data) > 2:
                    curr_data = curr_data[:2]
                data.append(list(map(float, curr_data)))

    # Reshape data
    data = np.concatenate([np.array(i) for i in data])
    data = (data[0::2] + 1j * data[1::2]).astype(complex)
    data = data.reshape((signal_index, -1)).T.squeeze()
    data = data.conj()

    return data, header


def num(s):
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def check_transmitter_frequency_in_header(header):
    """Check that the central frequency in the header is in MHz.
    Correct if not.

    :param header: Header dict containing 'TransmitterFrequency' key.
    :type header: dict
    :return: Checked and modified header
    :rtype: dict
    """
    if header['TransmitterFrequency'] > 1E5:
        header['TransmitterFrequency'] /= 1E6

    return header


def jmrui_hdr_to_obj(header):
    """Translate jMRUI txt header to NIfTI MRS"""

    header = check_transmitter_frequency_in_header(header)

    if 'TypeOfNucleus' in header:
        nucleus = id_nucleus(header['TypeOfNucleus'],
                             header['TransmitterFrequency'],
                             header['MagneticField'])
    else:
        nucleus = id_nucleus(None,
                             header['TransmitterFrequency'],
                             header['MagneticField'])

    meta = Hdr_Ext(
        float(header['TransmitterFrequency']),
        nucleus)

    if 'Spectrometer' in header:
        meta.set_standard_def('ManufacturersModelName', header['Spectrometer'])
    if 'NameOfPatient' in header:
        meta.set_standard_def('PatientName', header['NameOfPatient'])

    meta.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)

    if 'AdditionalInformation' in header:
        meta.set_user_def(key='AdditionalInformation',
                          value=header['AdditionalInformation'],
                          doc='jMRUI AdditionalInformation field.')
    if 'SignalNames' in header:
        meta.set_user_def(key='SignalNames',
                          value=header['SignalNames'],
                          doc='jMRUI SignalNames field.')

    return meta
